from tcutility.data import functionals
import json


with open('available_functionals.json', 'w+') as outp:
    for functional, functional_info in functionals.get_available_functionals().items():
        outp.write(json.dumps(functional_info, indent=4))
        outp.write('\n\n')



ensure_list = lambda x: [x] if not isinstance(x, (list, tuple, set)) else list(x)  # noqa: E731
squeeze_list = lambda x: x[0] if len(x) == 1 else x  # noqa: E731


def ensure_2d(x, transposed=False):
    x = ensure_list(x)
    if transposed:
        if not isinstance(x[0], (list, tuple, set)):
            x = [ensure_list(x)]
        else:
            x = [ensure_list(y) for y in x]
    else:
        x = [ensure_list(y) for y in x]
    return x


from tcutility import job, constants, formula, geometry, log, molecule, results, slurm, data, analysis, report, pathfunc  # noqa: F401, E402



from tcutility.data import functionals


def test_get():
	assert functionals.get('BLYP-D3(BJ)').name == 'BLYP-D3(BJ)'
	assert functionals.get('BLYP-D3(BJ)').path_safe_name == 'BLYP-D3BJ'
	assert functionals.get('BLYP-D3(BJ)').name_no_disp == 'BLYP'
	assert functionals.get('BLYP-D3(BJ)').category == 'GGA'
	assert functionals.get('BLYP-D3(BJ)').dispersion == 'D3(BJ)'
	assert functionals.get('BLYP-D3(BJ)').dispersion_name == 'GRIMME3 BJDAMP'
	assert functionals.get('BLYP-D3(BJ)').includes_disp is False
	assert functionals.get('BLYP-D3(BJ)').use_libxc is False
	assert functionals.get('BLYP-D3(BJ)').available_in_adf is True


def test_get2():
	assert functionals.get('rev-DSD-PBEP86').name == 'rev-DSD-PBEP86'
	assert functionals.get('rev-DSD-PBEP86').name_no_disp == 'rev-DSD-PBEP86'
	assert functionals.get('rev-DSD-PBEP86').category == 'DoubleHybrid'
	assert functionals.get('rev-DSD-PBEP86').dispersion is None
	assert functionals.get('rev-DSD-PBEP86').dispersion_name is None
	assert functionals.get('rev-DSD-PBEP86').includes_disp is False
	assert functionals.get('rev-DSD-PBEP86').use_libxc is False
	assert functionals.get('rev-DSD-PBEP86').available_in_adf is True


def test_functional_name_from_path_safe_name():
	assert functionals.functional_name_from_path_safe_name('BLYPD3BJ') == 'BLYP-D3(BJ)'

def test_functional_name_from_path_safe_name2():
	assert functionals.functional_name_from_path_safe_name('REVDSDPBEP86') == 'rev-DSD-PBEP86'

def test_functional_name_from_path_safe_name3():
	for func, info in functionals.get_available_functionals().items():
		assert functionals.functional_name_from_path_safe_name(info.path_safe_name) == func

def test_functional_name_from_path_safe_name4():
	for func, info in functionals.get_available_functionals().items():
		assert functionals.functional_name_from_path_safe_name(info.path_safe_name.replace("-", "")) == func

if __name__ == '__main__':
	import pytest 

	pytest.main()



import subprocess as sp
import os

fonts = [
	'calibri',
	'arial',
	'times new roman',
	'helvetica',
	'garamond',
]

characters = 'abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789⁰¹²³⁴⁵⁶⁷⁸⁹⁺⁻⁼⁽⁾ⁿⁱ₀₁₂₃₄₅₆₇₈₉₊₋₌₍₎ₐₑₒₓₔαβγδεζηθικλμνξοπρστυφχψωΑΒΓΔΕΖΗΘΙΚΛΜΝΞΟΠΡΣΤΥΦΧΨΩ.,/?<>{}[]()+-_=^:;%* '

script = """
const jsdom = require("jsdom");
const { JSDOM } = jsdom;

global.document = new JSDOM("<html><head></head></html>").window.document;

function getTextWidth(text, font) {
  var canvas = getTextWidth.canvas || (getTextWidth.canvas = global.document.createElement("canvas"));
  var context = canvas.getContext("2d");
  context.font = font;
  var metrics = context.measureText(text);
  return metrics.width;
};

let chars = '[CHARACTERS]';
let fonts = [[FONTS]];
for (let j = 0; j < fonts.length; j++) {
  let font = fonts[j];
  for (let i = 0; i < chars.length; i++) {
    console.log(font.replaceAll(" ", "_") + " " + chars[i] + " " + getTextWidth(chars[i], "1pt " + font));
  }
}
""".replace('[CHARACTERS]', characters).replace('[FONTS]', ", ".join(['"' + font + '"' for font in fonts]))

with open('char_width.js', 'w+') as js_script:
	js_script.write(script)

with open('_char_widths.txt', 'w+') as out:
	for line in sp.check_output(['node', 'char_width.js']).decode().splitlines():
		out.write(line + '\n')

os.remove('char_width.js')



'''
This module is used for resizing pictures containing molecules.
'''
import numpy as np
import cv2
import matplotlib.pyplot as plt
import os
from typing import Optional, Dict, Union


def _analyse_img(file, plot=False):
    '''
    Function used for analysing and getting useful information from an image.
    This includes circle locations and sizes.
    '''
    # Read image
    img = cv2.imread(file, cv2.IMREAD_UNCHANGED)
    img = _remove_padding(img)
    # Smooth it
    img_copy = img.copy()

    # Convert to greyscale
    img_gray = cv2.medianBlur(img_copy, 3)
    img_gray = cv2.cvtColor(img_gray, cv2.COLOR_BGRA2GRAY)

    # Apply Hough transform to greyscale image
    circles = cv2.HoughCircles(img_gray, cv2.HOUGH_GRADIENT, 1, 100,
                               param1=60, param2=40, minRadius=0, maxRadius=200)
    circles = np.uint16(np.around(circles))[0]
    circles = circles[(-circles[:, 2]).argsort()]

    if plot:
        img_copy = cv2.cvtColor(img_copy, cv2.COLOR_BGR2RGB)
        for i, circle in enumerate(circles):
            # draw the outer circle
            cv2.circle(img_copy,(circle[0],circle[1]),circle[2],(0,255,0),2)
            # draw the center of the circle
            cv2.circle(img_copy,(circle[0],circle[1]),2,(0,0,255),3)
            cv2.putText(img_copy, str(i), (circle[0] - 40, circle[1] + 44), cv2.FONT_HERSHEY_SIMPLEX, 4, (0, 0, 0), 16, cv2.LINE_AA)
            cv2.putText(img_copy, str(i), (circle[0] - 40, circle[1] + 44), cv2.FONT_HERSHEY_SIMPLEX, 4, (255, 255, 255), 8, cv2.LINE_AA)

        plt.imshow(img_copy)
        plt.draw()
        plt.pause(0.001)
    return circles, img


def _remove_padding(img):
    '''
    Function used to remove padding from an image.
    '''
    gray = cv2.cvtColor(img, cv2.COLOR_BGRA2GRAY)
    gray = 255*(gray < 240).astype(np.uint8) # To invert the text to white
    coords = cv2.findNonZero(gray) # Find all non-zero points (text)
    x, y, w, h = cv2.boundingRect(coords) # Find minimum spanning bounding box
    rect = img[y:y+h, x:x+w] # Crop the image - note we do this on the original image
    return rect


def resize(d, circle_numbers: Optional[Dict] = None, padding: Union[str, int, float] = 0):
    '''
    The main function for this module.
    Takes a directory `d` and selected circles and resizes and moves images in order to produce new aligned images.

    Args:
        d: the directory to take images from.
        circle_numbers: a dictionary containing the image names as keys and circle indices as the values.
            The new images will be aligned based on these selected circles.
        padding: the amount of padding to add to the new images.
            If given an integer, it adds `padding` number of pixels.
            Alternatively a string consisting of a number and a `%` sign can be given to add relative padding.
            E.g. `padding='10%'` will add a 10% padding to the images.
    '''
    circles = {}
    imgs = {}
    for file in os.listdir(d):
        if file == '.DS_Store':
            continue
        if file not in circle_numbers:
            continue
        circles_, img_ = _analyse_img(os.path.join(d, file))
        circles[file] = circles_[circle_numbers[file]]
        imgs[file] = img_

    # get the atom sizes, which is simply the radius of the circle we selected
    atom_sizes = {file: circle[2] for file, circle in circles.items()}
    # calculate ratios for resizing the images
    ratios = {file: max(atom_sizes.values())/size for file, size in atom_sizes.items()}
    # resize old images to make new ones
    imgs = {file: cv2.resize(img, [int(round(img.shape[1] * ratios[file])), int(round(img.shape[0] * ratios[file]))]) for file, img in imgs.items()}
    # also resize circles to match the new ratio
    circles = {file: circle * ratios[file] for file, circle in circles.items()}

    # now find out where to place the circles
    circle_x = [circle[0] for circle in circles.values()]
    circle_y = [circle[1] for file, circle in circles.items()]

    pad_left = {file: int((max(circle_x) - circle[0])) for file, circle in circles.items()}
    pad_top = {file: int((max(circle_y) - circle[1])) for file, circle in circles.items()}

    pad = {file: [(pad_top[file], 0), (pad_left[file], 0), (0, 0)] for file in imgs}
    imgs = {file: np.pad(img, pad[file], constant_values=0) for file, img in imgs.items()}

    imsize_x = [img.shape[1] for img in imgs.values()]
    imsize_y = [img.shape[0] for img in imgs.values()]

    pad_right = {file: int(max(imsize_x) - img.shape[1]) for file, img in imgs.items()}
    pad_bottom = {file: int(max(imsize_y) - img.shape[0]) for file, img in imgs.items()}

    pad = {file: [(0, pad_bottom[file]), (0, pad_right[file]), (0, 0)] for file in imgs}
    imgs = {file: np.pad(img, pad[file], constant_values=0) for file, img in imgs.items()}

    # add final padding
    if isinstance(padding, str) and padding.endswith('%'):
        padding = float(padding.removesuffix('%'))/100
        padding = int(padding * list(imgs.values())[0].shape[1]), int(padding * list(imgs.values())[0].shape[0])
    else:
        padding = int(padding), int(padding)

    pad = {file: [(padding[0], padding[0]), (padding[1], padding[1]), (0, 0)] for file in imgs}
    imgs = {file: np.pad(img, pad[file], constant_values=0) for file, img in imgs.items()}

    os.makedirs(d + '_fixed', exist_ok=True)
    for file, img in imgs.items():
        cv2.imwrite(os.path.join(d + '_fixed', file), img)

    cv2.imwrite(os.path.join(d + '_fixed', 'empty.png'), np.zeros_like(img) + np.array([255, 255, 255, 0]))
    ret = {file: os.path.join(d + '_fixed', file) for file in imgs}
    ret['empty.png'] = os.path.join(d + '_fixed', 'empty.png')
    return ret



from __future__ import annotations

import copy
import functools
import pathlib as pl
from itertools import zip_longest
from typing import Dict, List, Optional, Sequence, Union

import attrs
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd

import tcutility.log as log
from tcutility.analysis.vdd import charge
from tcutility.constants import VDD_UNITS
from tcutility.results import result

PRINT_FORMAT = {"me": "%+.0f", "e": "%+.3f"}


def change_unit_decorator(func):
    @functools.wraps(func)
    def wrapper(self, unit: str = "me", *args, **kwargs):
        current_unit = self.unit
        self.change_unit(unit)
        result = func(self, *args, **kwargs)
        self.change_unit(current_unit)
        return result

    return wrapper


def create_vdd_charge_manager(results: result.Result) -> VDDChargeManager:
    """Create a VDDChargeManager from a Result object."""
    vdd_charges: Dict[str, List[charge.VDDCharge]] = {}
    atom_symbols = results.molecule.atom_symbols  # type: ignore
    frag_indices = results.molecule.frag_indices  # type: ignore
    calc_dir = pl.Path(results.files["root"])  # type: ignore
    is_fragment_calculation = results.adf.used_regions  # type: ignore
    mol_charge = results.molecule.mol_charge  # type: ignore

    # Convert the VDD charges to VDDCharge objects
    for irrep, charge_array in results.properties.vdd.items():  # type: ignore
        irrep = "vdd" if irrep == "charges" else irrep
        vdd_charges[irrep] = []
        for atom_index, (frag_index, vdd_charge) in enumerate(zip(frag_indices, charge_array)):
            atom_symbol = atom_symbols[atom_index]
            vdd_charges[irrep].append(charge.VDDCharge(atom_index + 1, atom_symbol, vdd_charge, frag_index))

    return VDDChargeManager(vdd_charges, is_fragment_calculation, calc_dir, mol_charge)


@attrs.define
class VDDChargeManager:
    """Class to manage the VDD charges. It can be used to print the VDD charges in a nice table and write them to a text file or excel file."""

    vdd_charges: Dict[str, List[charge.VDDCharge]]  # {"vdd": [VDDCharge, ...], "irrep1": [VDDCharge, ...], ...}
    is_fragment_calculation: bool
    calc_dir: pl.Path
    mol_charge: int
    name: str = attrs.field(init=False)
    unit: str = attrs.field(init=False, default="e")  # unit of the VDD charges. Available units are "me" (mili-electrons) and "e" (electrons)
    irreps: List[str] = attrs.field(init=False, default=set())

    def __attrs_post_init__(self):
        self.irreps = list(self.vdd_charges.keys())
        self.name = self.calc_dir.name if self.calc_dir is not None else self.name
        self.change_unit("me")
        log.info(f"VDDChargeManager created for {self.name}")

    def __str__(self) -> str:
        """Prints the VDD charges in a nice table. Checks if the calculation is a fragment calculation and prints the summed VDD charges if it is."""
        individual_charges_table = self.get_vdd_charges_table()
        summed_charges_table = self.get_summed_vdd_charges_table()

        ret_str = f"{self.name}\nVDD charges (in unit {self.unit}):\n{individual_charges_table}\n\n"
        if self.is_fragment_calculation:
            ret_str += f"Summed VDD charges (in unit {self.unit}):\n{summed_charges_table}\n"

        return ret_str

    def charge_is_conserved(self) -> bool:
        """Check if the total charge of the molecule is conserved. The total charge is the sum of the VDD charges."""
        tolerance = 1e-4 if self.unit == "e" else 1e-1
        is_conserved = np.isclose(self.mol_charge, sum([charge.charge for charge in self.vdd_charges["vdd"]]), atol=tolerance)
        return is_conserved  # type: ignore since numpy _bool is not recognized as bool

    def change_unit(self, new_unit: str) -> None:
        """Change the unit of the VDD charges. Available units are "me" (mili-electrons) and "e" (electrons)."""
        if new_unit not in VDD_UNITS:
            raise ValueError(f"Unit {new_unit} is not available. Choose from {VDD_UNITS.keys()}")

        if new_unit == self.unit:
            return

        ratio = VDD_UNITS[new_unit] / VDD_UNITS[self.unit]
        [charge.change_unit(ratio) for charges in self.vdd_charges.values() for charge in charges]
        self.unit = new_unit

    @change_unit_decorator
    def get_vdd_charges(self) -> Dict[str, List[charge.VDDCharge]]:
        """Get the VDD charges in the specified unit ([me] or [e])."""
        return copy.deepcopy(self.vdd_charges)

    @change_unit_decorator
    def get_summed_vdd_charges(self, irreps: Optional[Sequence[str]] = None) -> Dict[str, Dict[str, float]]:
        """Get the summed VDD charges per fragment for the specified unit ([me] or [e])."""
        irreps = irreps if irreps is not None else self.irreps
        summed_vdd_charges: Dict[str, Dict[str, float]] = {}

        for irrep in irreps:
            summed_vdd_charges[irrep] = {}
            for vdd_charge in self.vdd_charges[irrep]:
                frag_index = str(vdd_charge.frag_index)
                summed_vdd_charges[irrep].setdefault(frag_index, 0.0)
                summed_vdd_charges[irrep][frag_index] += vdd_charge.charge

        return copy.deepcopy(summed_vdd_charges)

    def get_vdd_charges_dataframe(self) -> pd.DataFrame:
        """Get the VDD charges as a pandas DataFrame in a specified unit ([me] or [e])."""
        frag_indices = [charge.frag_index for charge in self.vdd_charges["vdd"]]
        atom_symbols = [f"{charge.atom_index}{charge.atom_symbol}" for charge in self.vdd_charges["vdd"]]
        charges = [[charge.charge for charge in charges] for _, charges in self.vdd_charges.items()]
        headers = ["Frag", "Atom"] + [irrep for irrep, _ in self.vdd_charges.items()]
        combined_table = list(zip_longest(frag_indices, atom_symbols, *charges, fillvalue=""))
        df = pd.DataFrame(combined_table, columns=headers).rename(columns={"vdd": "Total"})
        return df

    def get_summed_vdd_charges_dataframe(self) -> pd.DataFrame:
        """Get the summed VDD charges as a pandas DataFrame in a specified unit ([me] or [e])."""
        summed_data = self.get_summed_vdd_charges()
        summed_data["Frag"] = {str(key): int(key) for key in summed_data["vdd"].keys()}
        df = pd.DataFrame(summed_data).pipe(lambda df: df[df.columns.tolist()[-1:] + df.columns.tolist()[:-1]])  # move the "Frag" column to the front
        return df.rename(columns={"vdd": "Total"})

    def get_vdd_charges_table(self) -> str:
        df = self.get_vdd_charges_dataframe()
        return df.to_string(float_format=lambda x: PRINT_FORMAT[self.unit] % x, justify="center", index=False, col_space=6)

    def get_summed_vdd_charges_table(self) -> str:
        df = self.get_summed_vdd_charges_dataframe()
        return df.to_string(float_format=lambda x: PRINT_FORMAT[self.unit] % x, justify="center", col_space=6, index=False)

    @staticmethod
    def write_to_txt(output_dir: Union[str, pl.Path], managers: Union[VDDChargeManager, Sequence[VDDChargeManager]], unit: str = "me") -> None:
        """Write the VDD charges to a text file. It is a static method because multiple managers can be written to the same file."""
        out_dir = pl.Path(output_dir) if not isinstance(output_dir, pl.Path) else output_dir
        files = [out_dir / "VDD_charges_per_atom.txt", out_dir / "VDD_charges_per_fragment.txt"]
        managers = managers if isinstance(managers, Sequence) else [managers]

        # Print charges per atom and per fragment in seperate files
        for i, file in enumerate(files):
            with open(file, "w") as f:
                f.write("VDD charges:\n")
                for manager in managers:
                    f.write(f"{manager.name} (unit = {manager.unit})\n")
                    if i == 0:
                        f.write(manager.get_vdd_charges_table())
                    else:
                        f.write(manager.get_summed_vdd_charges_table()) if manager.is_fragment_calculation else f.write("No fragment calculation\n")
                    f.write("\n\n")

    def write_to_excel(self, output_file: Optional[Union[str, pl.Path]] = None) -> None:
        """Write the VDD charges to an excel file. Results are written to two sheets: "VDD charges" and "Summed VDD charges"."""
        file = pl.Path(output_file) if output_file is not None else self.calc_dir / f"vdd_charges_{self.name}.xlsx"
        file = file.with_suffix(".xlsx") if not file.suffix == ".xlsx" else file

        df = self.get_vdd_charges_dataframe()

        with pd.ExcelWriter(file) as writer:
            df.to_excel(writer, sheet_name=f"VDD charges (in {self.unit})", index=False, float_format=PRINT_FORMAT[self.unit])
            if self.is_fragment_calculation:
                df_summed = self.get_summed_vdd_charges_dataframe()
                df_summed.to_excel(writer, sheet_name=f"Summed VDD charges  (in {self.unit})", index=False, float_format=PRINT_FORMAT[self.unit])

    def plot_vdd_charges_per_atom(self, output_file: Optional[Union[str, pl.Path]] = None, unit: str = "me") -> None:
        """Plot the VDD charges as a bar graph for each irrep."""
        file = pl.Path(output_file) if output_file is not None else self.calc_dir / f"vdd_charges_{self.name}.png"
        file = file.with_suffix(".png") if not file.suffix == ".png" else file
        self.change_unit(unit)

        # Increase the global font size
        plt.rcParams.update({"font.size": 14})

        num_irreps = len(self.vdd_charges)
        n_max_charges = max([len(charges) for charges in self.vdd_charges.values()])
        _, axs = plt.subplots(num_irreps, 1, figsize=(n_max_charges * 1.15, 5 * num_irreps), sharey=True)
        axs = [axs] if num_irreps == 1 else axs

        # Initialize a variable to keep track of the most positive/negative values
        adjusted_abs_max_values = []

        counter = 1
        for ax, (irrep, charges) in zip(axs, self.vdd_charges.items()):
            atom_symbols = [f"{charge.atom_index}{charge.atom_symbol} ({charge.frag_index})" for charge in charges]
            charge_values = [charge.charge for charge in charges]

            bars = ax.bar(atom_symbols, charge_values, color="sandybrown", edgecolor="black")
            if counter == len(axs):
                ax.set_xlabel("Atom (#Fragment Number)")
            ax.set_ylabel(f"Charge ({unit})")
            ax.set_title(f"VDD Charges {self.name} - {irrep}")
            ax.yaxis.grid(True)  # Only display vertical grid lines

            # Add the charge value on top of each bar. If the value is negative, place the text below the bar
            for bar in bars:
                yval = bar.get_height()
                x_pos = bar.get_x() + bar.get_width() / 2

                if yval <= 0:
                    ax.text(x_pos, yval + yval * 0.1, int(yval), va="top", ha="center")
                else:
                    ax.text(x_pos, yval + yval * 0.1, int(yval), va="bottom", ha="center")
                adjusted_abs_max_values.append(yval + yval * 0.1)
            counter += 1

        # Making sure the values do not extend outside the plot. Also determines the y-axis limits for all subplotss
        min_value, max_value = min(adjusted_abs_max_values), max(adjusted_abs_max_values)

        # Axes adjustments
        # Set the y-axis limits for all subplots based on the most positive/negative values
        for ax in axs:
            ax.set_ylim([min_value - abs(min_value) * 0.2, max_value + abs(max_value) * 0.2])

        # plt.tight_layout()
        plt.savefig(file, dpi=300)



import time
import functools
import json
import os
import platformdirs


_timed_func_cache = {}
_time_since_cache = {}

_func_cache = {}

_general_cache = {}

_cache_dir = platformdirs.user_cache_dir(appname='TCutility', appauthor='TheoCheM', ensure_exists=True)

def timed_cache(delay: float):
    '''
    Decorator that creates a timed cache for the function or method. 
    This cache will expire after a chosen amount of time.

    Args:
        delay: the expiry time for the function cache.
    '''
    def decorator(func):
        # this is the main decorator returned by timed_cache
        @functools.wraps(func)
        def inner_decorator(*args, **kwargs):
            # we have to create a tuple of the kwargs items to ensure we can hash the arguments
            arguments = args, tuple(kwargs.items())

            # the time since the last caching
            dT = time.perf_counter() - _time_since_cache[func]
            # if the cache has expired, we remove it from the cache
            # and set the new time_since_cache
            if dT >= delay:
                _timed_func_cache[func].pop(arguments, None)
                _time_since_cache[func] = time.perf_counter()

            # check if the arguments were called before
            if arguments in _timed_func_cache[func]:
                return _timed_func_cache[func][arguments]

            # if it is not present we add it to the cache
            res = func(*args, **kwargs)
            _timed_func_cache[func][arguments] = res
            return res

        # initialize the cache and time_since_cache
        _timed_func_cache[func] = {}
        _time_since_cache[func] = -delay

        return inner_decorator
    return decorator


def cache(func):
    '''
    Function decorator that stores results from previous calls to the function or method.
    '''
    @functools.wraps(func)
    def inner_decorator(*args, **kwargs):
        # we have to create a tuple of the kwargs items to ensure we can hash the arguments
        arguments = args, tuple(kwargs.items())

        # check if the arguments were called before
        if arguments in _func_cache[func]:
            return _func_cache[func][arguments]

        # if it is not present we add it to the cache
        res = func(*args, **kwargs)
        _func_cache[func][arguments] = res
        return res

    _func_cache[func] = {}
    return inner_decorator


def _get_from_cache_file(file, func, args, kwargs):
    '''
    Retrieve results from a JSON file. Return `None` if not found.
    '''
    # open the file and parse the data
    with open(os.path.join(_cache_dir, file)) as cfile:
        data = json.loads(cfile.read())

    # we now go through the data
    # it is formatted as a list of dicts
    # each dict has the 'func', 'args', 'kwargs' and 'value' keys
    for datum in data:
        # func is set as the function qualname
        if datum['func'] != func.__qualname__:
            continue

        # args is retrieved as a list
        if datum['args'] != list(args):
            continue

        # kwargs is simply a dict
        if datum['kwargs'] != kwargs:
            continue

        # if we did not exit the loop yet we return the value
        return datum['value']


def _write_to_cache_file(file, func, args, kwargs, value):
    '''
    Write results to the file.
    '''
    # we open the file to get the data
    with open(os.path.join(_cache_dir, file)) as cfile:
        data = json.loads(cfile.read())

    # add the new results to the file
    new = {
        'func': func.__qualname__,
        'args': args,
        'kwargs': kwargs,
        'value': value
    }
    data.append(new)

    with open(os.path.join(_cache_dir, file), 'w+') as cfile:
        cfile.write(json.dumps(data, indent=4))


def _clear_cache_file(file):
    '''
    Function that clears a file and writes a new beginning of a list.
    '''
    os.makedirs(_cache_dir, exist_ok=True)
    with open(os.path.join(_cache_dir, file), 'w+') as cfile:
        cfile.write('[]')


def cache_file(file):
    '''
    Function decorator that stores results of a function to a file.
    Because results are written to a file the values persist between Python sessions.
    This is useful, for example, for online API calls.

    Args:
        file: the filepath to store function call results to. 
            Files will be stored in the platform dependent temporary file directory.

    .. seealso::
        `platformdirs.user_cache_dir <https://platformdirs.readthedocs.io/en/latest/api.html#platformdirs.user_cache_dir>`_ for information on the temporary directory.
    '''
    def decorator(func):
        # make the file if it doesnt exist
        if not os.path.exists(os.path.join(_cache_dir, file)):
            _clear_cache_file(file)

        @functools.wraps(func)
        def inner_decorator(*args, **kwargs):
            # check if the arguments were called before
            cached = _get_from_cache_file(file, func,  args, kwargs)
            if cached is not None:
                return cached

            # if it is not present we add it to the cache file
            res = func(*args, **kwargs)
            _write_to_cache_file(file, func,  args, kwargs, res)
            return res

        return inner_decorator
    return decorator



import os

import numpy as np

from tcutility import constants, molecule
from tcutility.results import Result

j = os.path.join


def get_calc_files(calc_dir: str) -> Result:
    """Function that returns files relevant to AMS calculations stored in ``calc_dir``.

    Args:
        calc_dir: path pointing to the desired calculation

    Returns:
        Result object containing filenames and paths
    """
    # collect all files in the current directory and subdirectories
    files = []
    for root, _, files_ in os.walk(calc_dir):
        files.extend([j(root, file) for file in files_])

    # parse the filenames
    ret = Result()
    ret.extra = []
    ret.root = os.path.abspath(calc_dir)
    for file in files:
        if file.endswith(".out") and not file.endswith("g98.out"):
            ret.out = os.path.abspath(file)
            continue
        if file.endswith("xtbopt.xyz"):
            ret.opt_out = os.path.abspath(file)
            continue
        if file.endswith("xtbopt.log"):
            ret.opt_history = os.path.abspath(file)
            continue
        if file.endswith("xtbscan.log"):
            ret.scan_out = os.path.abspath(file)
            continue
        if file.endswith("vibspectrum"):
            ret.vibspectrum = os.path.abspath(file)
            continue
        if file.endswith("hessian"):
            ret.hessian = os.path.abspath(file)
            continue

        ret.extra.append(os.path.abspath(file))

    return ret


def get_version(info: Result) -> Result:
    """Function to get the ORCA version used in the calculation.

    Args:
        info: Result object containing ORCA calculation settings.

    Returns:
        :Result object containing results about the ORCA version:

            - **full (str)** – the full version string as written by ORCA.
            - **major (str)** – major ORCA version.
            - **minor (str)** – minor ORCA version.
            - **micro (str)** – micro ORCA version.
    """
    ret = Result()
    with open(info.files.out) as out:
        for line in out.readlines():
            line = line.strip()
            if "* xtb version" not in line:
                continue
            version = line.split()[3]
            ret.full = version
            ret.major = version.split(".")[0]
            ret.minor = version.split(".")[1]
            ret.micro = version.split(".")[2]
            return ret


def get_input(info: Result) -> Result:
    """Function that parses the input file for this ORCA calculation.

    Args:
        info: Result object containing ORCA calculation settings.

    Returns:
        :Result object containing information about the calculation input:

            - **main (list[str])** - the main inputs for the calculation. These are the lines that start with a "!".
            - **sections (Result)** - extra settings added to the calculation. These are the lines that start with a "%" and optionally end with "END" clause.
            - **system (Result)** - settings related to the molecular system. This includes charge, multiplicity and the coordinates.
            - **task (str)** - the task that was performed by the calculation, e.g. "SinglePoint", "TransitionStateSearch".
    """
    ret = Result()
    with open(info.files.out) as out:
        for line in out.readlines():
            line = line.strip()
            if line.startswith("program call"):
                call = line.split(":")[1].strip().split()
            if line.startswith("coordinate file"):
                ret.coord_file = line.split(":")[1].strip()

    ret.call = " ".join(call)

    ### TASK
    ret.task = "SinglePoint"
    for option in ["--opt", "--ohess"]:
        # check if we used the option in our call
        if option not in call:
            continue

        # if we used it we did a geo-opt
        ret.task = "GeometrOptimization"
        # check if we did a PESScan, it also requires using the --opt option
        if "scan_out" in info.files:
            ret.task = "PESScan"

        # check if we gave convergence criteria
        ret.geometry_convergence = "Normal"
        option_idx = call.index(option)
        if not call[option_idx + 1].startswith("-"):
            ret.geometry_convergence = call[option_idx + 1]

    ### CHARGE
    ret.charge = 0
    for option in ["--charge", "-c"]:
        # check if we used the option in our call
        if option not in call:
            continue
        # read the next position in the call
        option_idx = call.index(option)
        ret.charge = int(call[option_idx + 1])

    ### SPIN-POLARIZATION
    ret.spin_polarization = 0
    for option in ["--uhf", "-u"]:
        # check if we used the option in our call
        if option not in call:
            continue
        # read the next position in the call
        option_idx = call.index(option)
        ret.spin_polarization = int(call[option_idx + 1])

    ### SOLVATION
    ret.solvent = None
    ret.solvation_model = None

    if "--alpb" in call:
        option_idx = call.index("--alpb")
        ret.solvent = call[option_idx + 1]
        ret.solvation_model = "ALPB"

    for option in ["-g", "--gbsa"]:
        if option not in call:
            continue
        option_idx = call.index(option)
        ret.solvent = call[option_idx + 1]
        ret.solvation_model = "GBSA"

    ### DETAILED INPUT
    ret.detailed = None
    if "--input" in call:
        ret.detailed = Result()
        option_idx = call.index("--input")
        inp_file = call[option_idx + 1]
        for file in info.files.extra:
            if file.endswith(inp_file):
                ret.detailed.file = file
                break

        with open(ret.detailed.file) as infile:
            content = infile.readlines()

        for line in content:
            line = line.strip()
            if line.startswith("$"):
                curr_section = line[1:]
                continue

            if ":" in line:
                key, val = line.split(":")
                ret.detailed[curr_section].setdefault(key, [])
                ret.detailed[curr_section][key].append(val.strip())

            elif "=" in line:
                key, val = line.split("=")
                ret.detailed[curr_section][key] = val.strip()

    ### MODEL HAMILTONIAN
    ret.model = "GFN2-xTB"
    if "--gfn" in call:
        option_idx = call.index("--gfn")
        version = call[option_idx + 1]
        ret.model = f"GFN{version}-xTB"

    if "--gfnff" in call:
        ret.model = "GFNFF"

    return ret


def get_calculation_status(info: Result) -> Result:
    """Function that returns the status of the ORCA calculation described in reader. In case of non-succes it will also give possible reasons for the errors/warnings.

    Args:
        info: Result object containing ORCA calculation information.

    Returns:
        :Result object containing information about the calculation status:

            - **fatal (bool)** – True if calculation cannot be analysed correctly, False otherwise
            - **reasons (list[str])** – list of reasons to explain the status, they can be errors, warnings, etc.
            - **name (str)** – calculation status written as a string, one of ("SUCCESS", "RUNNING", "UNKNOWN", "SUCCESS(W)", "FAILED")
            - **code (str)** – calculation status written as a single character, one of ("S", "R", "U", "W" "F")
    """
    ret = Result()
    ret.fatal = True
    ret.name = None
    ret.code = None
    ret.reasons = []

    if "out" not in info.files:
        ret.reasons.append("Calculation status unknown")
        ret.name = "UNKNOWN"
        ret.code = "U"
        return ret

    with open(info.files.out) as out:
        lines = out.readlines()
        if any(["[WARNING] Runtime exception occurred" in line for line in lines]):
            ret.fatal = True
            ret.name = "FAILED"
            ret.code = "F"

            line_index = [i for i, line in enumerate(lines) if "[WARNING] Runtime exception occurred" in line][0]
            for line in lines[line_index + 1 :]:
                if "##########" in line:
                    break
                ret.reasons.append(line.strip())
            return ret

        if any(["* finished run" in line for line in lines]):
            ret.fatal = False
            ret.name = "SUCCESS"
            ret.code = "S"
            return ret

    ret.name = "FAILED"
    ret.code = "F"
    return ret


def get_molecules(info: Result) -> Result:
    """Function that returns information about the molecules for this calculation.

    Args:
        info: Result object containing ORCA calculation information.

    Returns:
        :Result object containing properties from the ORCA calculation:

            - **input (plams.Molecule)** - the input molecule for this calculation.
            - **output (plams.Molecule)** - the output molecule for this calculation, for example the optimized structure after a geometry optimization.
            - **number_of_atoms (int)** - the number of atoms in the molecular system.
    """
    ret = Result()
    for file in info.files.extra:
        if file.endswith(info.input.coord_file):
            coord_file = file
            break

    ret.input = molecule.load(coord_file)
    ret.output = ret.input.copy()

    if "opt_out" in info.files:
        ret.output = molecule.load(info.files.opt_out)

    return ret


def get_info(calc_dir: str) -> Result:
    """Function to read useful info about the calculation in ``calc_dir``. Returned information will depend on the type of file that is provided.

    Args:
        calc_dir: path pointing to the desired calculation.

    Returns:
        :Result object containing results about the calculation and AMS:

            - **version (Result)** – information about the AMS version used, see :func:`get_version`.
            - **files (Result)** - paths to files that are important for this calculation.
            - **input (Result)** - information about the input of this calculation, see :func:`get_input`.
            - **level (Result)** - information about the level of theory used for this calculation, see :func:`get_level_of_theory`.
            - **engine (str)** – the engine that was used to perform the calculation, for example 'adf', 'dftb', ...
            - **status (Result)** – information about calculation status, see :func:`get_calculation_status`.
            - **molecule (Result)** – information about the input and output molecules and the molecular system in general, see :func:`get_molecules`.
    """
    ret = Result()

    ret.engine = "xtb"
    ret.files = get_calc_files(calc_dir)

    # store the input of the calculation
    ret.input = get_input(ret)

    # store information about the version of AMS
    ret.version = get_version(ret)

    # # store the calculation status
    ret.status = get_calculation_status(ret)

    # # read molecules
    ret.molecule = get_molecules(ret)

    return ret


def get_properties(info: Result) -> Result:
    """Function to get properties from an ORCA calculation.

    Args:
        info: Result object containing ORCA properties.

    Returns:
        :Result object containing properties from the ORCA calculation:

            - **energy.bond (float)** – total bonding energy (|kcal/mol|).
            - **energy.enthalpy (float)** – enthalpy (|kcal/mol|). Only obtained if vibrational modes were calculated.
            - **energy.entropy (float)** – entropy (|kcal/mol|). Only obtained if vibrational modes were calculated.
            - **energy.gibbs (float)** – Gibb's free energy (|kcal/mol|). Only obtained if vibrational modes were calculated.
            - **energy.[method] (float)** - total energy (|kcal/mol|) at a certain level (HF, MP2, CCSD, ...). This is the sum of energy.HF and energy.[method]_corr.
            - **energy.[method]_corr (float)** - electron correlation energy (|kcal/mol|) at a certain level (HF, MP2, CCSD, ...).
            - **vibrations.number_of_modes (int)** – number of vibrational modes for this molecule, 3N-5 for non-linear molecules and 3N-6 for linear molecules, where N is the number of atoms.
            - **vibrations.number_of_imaginary_modes (int)** – number of imaginary vibrational modes for this molecule.
            - **vibrations.frequencies (list[float])** – vibrational frequencies associated with the vibrational modes, sorted from low to high (|cm-1|).
            - **vibrations.intensities (list[float])** – vibrational intensities associated with the vibrational modes (|km/mol|). In ORCA, intensities of imaginary modes are set to 0.
            - **vibrations.modes (list[float])** – list of vibrational modes sorted from low frequency to high frequency.
            - **vibrations.character (str)** – the PES character of the molecular system. Either "minimum", "transitionstate" or "saddlepoint(n_imag)", for 0, 1, n_imag number of imaginary frequencies.
            - **t1** - T1 diagnostic for the highest level of correlation chosen. Used to check the validity of the reference wavefunction.
            - **s2** - expectation value of the :math:`S^2` operator.
            - **s2_expected** - ideal expectation value of the :math:`S^2` operator.
            - **spin_contamination** - the amount of spin-contamination observed in this calculation. It is equal to (s2 - s2_expected) / (s2_expected). Ideally this value should be below 0.1.
    """
    ret = Result()

    with open(info.files.out) as out:
        lines = [line.strip() for line in out.readlines()]

    # read some important info about the calculation
    for line in lines:
        if "TOTAL ENERGY" in line:
            ret.energy.bond = float(line.split()[3]) * constants.HA2KCALMOL
            continue

        if "TOTAL FREE ENERGY" in line:
            ret.energy.gibbs = float(line.split()[4]) * constants.HA2KCALMOL
            continue

        if "TOTAL ENTHALPY" in line:
            ret.energy.enthalpy = float(line.split()[3]) * constants.HA2KCALMOL
            continue

        if "# frequencies" in line:
            ret.vibrations.number_of_modes = int(line.split()[3])
            continue

        if "# imaginary freq." in line:
            ret.vibrations.number_of_imaginary_modes = int(line.split()[4])
            continue

        if "vibrational frequencies" in line:
            ret.vibrations.frequencies = []
            continue

        if "eigval :" in line:
            ret.vibrations.frequencies.extend([float(freq) for freq in line.split()[2:]])

    if ret.vibrations.frequencies:
        # get the number of the first vibrational mode, so without the translations and rotations
        first_mode = len(ret.vibrations.frequencies) - ret.vibrations.number_of_modes + 1
        # we have to remove the first 5 or 6 frequencies, because they are translation+rotation
        ret.vibrations.frequencies = ret.vibrations.frequencies[first_mode - 1 :]

        # read in the vibspectrum file to get the vibrational intensities:
        ret.vibrations.intensities = []
        if info.files.vibspectrum:
            with open(info.files.vibspectrum) as spec:
                for line in spec.readlines()[3:-1]:
                    if line.startswith("$"):
                        continue

                    if int(line.split()[0]) < first_mode:
                        continue

                    ret.vibrations.intensities.append(float(line.split()[3]))

        # read in the hessian file to get the normal modes:
        if info.files.hessian:
            hessian = []
            with open(info.files.hessian) as hess:
                for line in hess.readlines()[1:]:
                    hessian.extend([float(x) for x in line.split()])

            # number of modes:
            N = int(np.sqrt(len(hessian)))
            # square hessian:
            H = np.reshape(hessian, (N, N))

            # get the atomic masses
            mol = molecule.load(info.files.opt_out)
            masses = np.atleast_2d([atom.mass for atom in mol.atoms for _ in range(3)])

            # make the reduced mass matrix
            mu = np.sqrt(masses * masses.T)
            # and reduced mass Hessian
            F = H / mu

            # diagonalize it to get frequencies and modes
            freqs, modes = np.linalg.eigh(F)
            # print(modes)

            # from yviewer import viewer
            # mol.guess_bonds()
            # viewer.show(mol, molinfo=[{'normalmode': modes[0].reshape(-1, 3)}])

    return ret


if __name__ == "__main__":
    from tcutility import log

    ret = get_info("/Users/yumanhordijk/PhD/ganesh_project/calculations/AlCl3_xtb/opt_1")
    log.log(ret.files)
    props = get_properties(ret)
    print(props)

    # ret = get_info("/Users/yumanhordijk/PhD/ganesh_project/calculations/AlCl3_xtb/scan_1")
    # print(ret)



from tcutility.data import atom
import pytest


def test_parse():
    assert atom.parse_element("hydrogen") == 1


def test_parse2():
    assert atom.parse_element("C") == 6


def test_parse3():
    assert atom.parse_element(14) == 14


def test_parse4():
    assert atom.parse_element(1) == 1


def test_parse5():
    assert atom.parse_element("Carbon") == 6


def test_radius():
    assert atom.radius("Tungsten") == 1.62


def test_radius2():
    with pytest.raises(KeyError):
        atom.radius("FakeElement")


if __name__ == "__main__":
    pytest.main()



from tcutility import ensure_2d


def test_0d_1():
    assert ensure_2d(5) == [[5]]


def test_0d_2():
    assert ensure_2d(None) == [[None]]


def test_0d_3():
    assert ensure_2d("abcdefghijklmnopqrstuvwxyz") == [["abcdefghijklmnopqrstuvwxyz"]]


def test_1d_1():
    assert ensure_2d([5]) == [[5]]


def test_1d_2():
    assert ensure_2d([None]) == [[None]]


def test_1d_3():
    assert ensure_2d(["abcdefghijklmnopqrstuvwxyz"]) == [["abcdefghijklmnopqrstuvwxyz"]]


def test_1d_4():
    assert ensure_2d([1, 2, 3, 4, 5]) == [[1], [2], [3], [4], [5]]


def test_1d_5():
    assert ensure_2d((1, 2, 3, 4, 5)) == [[1], [2], [3], [4], [5]]


def test_1d_6():
    assert ensure_2d({1, 2, 3, 4, 5}) == [[1], [2], [3], [4], [5]]


def test_1d_7():
    assert ensure_2d([1, 2, 3, 4, 5], transposed=True) == [[1, 2, 3, 4, 5]]


def test_1d_8():
    assert ensure_2d((1, 2, 3, 4, 5), transposed=True) == [[1, 2, 3, 4, 5]]


def test_1d_9():
    assert ensure_2d({1, 2, 3, 4, 5}, transposed=True) == [[1, 2, 3, 4, 5]]


def test_2d_1():
    assert ensure_2d([[1, 2, 3, 4, 5]]) == [[1, 2, 3, 4, 5]]


def test_2d_2():
    assert ensure_2d(((1, 2, 3, 4, 5),)) == [[1, 2, 3, 4, 5]]


def test_2d_3():
    assert ensure_2d([{1, 2, 3, 4, 5}]) == [[1, 2, 3, 4, 5]]


def test_2d_4():
    assert ensure_2d([[1, 2, 3, 4, 5]], transposed=True) == [[1, 2, 3, 4, 5]]


def test_2d_5():
    assert ensure_2d(((1, 2, 3, 4, 5),), transposed=True) == [[1, 2, 3, 4, 5]]


def test_2d_6():
    assert ensure_2d([{1, 2, 3, 4, 5}], transposed=True) == [[1, 2, 3, 4, 5]]


if __name__ == "__main__":
    import pytest

    pytest.main()



import numpy as np
from math import sin, cos, atan2, sqrt
import scipy
from typing import Tuple, Union, Sequence
from scm import plams


class Transform:
    '''
    Transformation matrix that handles rotation, translation and scaling of sets of 3D coordinates.

    Build and return a transformation matrix.
    This 4x4 matrix encodes rotations, translations and scaling.
    
    :math:`\\textbf{M} = \\begin{bmatrix} \\textbf{R}\\text{diag}(S) & \\textbf{T} \\\\ \\textbf{0}_3 & 1 \\end{bmatrix}`

    where :math:`\\textbf{R} \\in \\mathbb{R}^{3 \\times 3}`, :math:`\\textbf{T} \\in \\mathbb{R}^{3 \\times 1}` and :math:`\\textbf{0}_3 = [0, 0, 0] \\in \\mathbb{R}^{1 \\times 3}`.

    When applied to a coordinates :math:`[\\textbf{x}, \\textbf{y}, \\textbf{z}, \\textbf{1}]^T \\in \\mathbb{R}^{n \\times 4}` it will apply these transformations simultaneously.
    '''
    def __init__(self):
        # initialize an empty transformation matrix
        self.M = self._build_matrix()

    def __str__(self):
        return str(self.M)

    def __call__(self, *args, **kwargs):
        return self.apply(*args, **kwargs)

    def apply(self, v: np.ndarray) -> np.ndarray:
        """
        Applies the transformation matrix to vector(s) :math:`v \\in \\mathbb{R}^{N \\times 3}`.

        Application is a three-step process:

        #. Append row vector of ones to the bottom of :math:`v`
        #. Apply the transformation matrix: :math:`\\textbf{M}v`
        #. Remove the bottom row vector of ones and return the result

        Returns:
            A new array :math:`v' = \\textbf{M}v` that has been transformed using this transformation matrix.

        .. note::

            The ``Transform.__call__()`` method redirects to this method. Calling ``transform.apply(coords)`` is the same as ``transform(coords)``.
        """
        is_mol = isinstance(v, plams.Molecule)
        if is_mol:
            mol = v.copy()

        v = np.array(v)
        v = np.atleast_2d(v)
        v = np.asarray(v).T
        N = v.shape[1]
        v = np.vstack([v, np.ones(N)])
        vprime = self.M @ v

        if is_mol:
            mol.from_array(vprime[:3, :].T)
            return mol

        return vprime[:3, :].T

    def __matmul__(self, other):
        if isinstance(other, Transform):
            return self.combine_transforms(other)

    def combine_transforms(self, other: "Transform") -> "Transform":
        """
        Combine two different transform objects. This involves creating
        a new Transform object and multiplying the two transform matrices
        and assigning it to the new object.

        Args:
            other: the transformation matrix object to combine this one with.

        Returns:
            A new transformation matrix that is a product of the original (left side) and other (right side) matrices.

        .. note::

            The ``Transform.__matmul__()`` method redirects to this method. Calling ``new = this.combine_transforms(other)`` is the same as ``new = this @ other``.
        """
        new = Transform()
        new.M = self.M @ other.M
        return new

    def translate(self, T: np.ndarray = None, x: float = None, y: float = None, z: float = None):
        """
        Add a translation component to the transformation matrix.
        Arguments can be given as a container of x, y, z values. They can also be given separately.
        You can also specify x, y and z components separately

        Example usage:
            | ``Transform.translate([2, 3, 0])``
            | ``Transform.translate(x=2, y=3)``
        """
        if T is None:
            T = [x or 0, y or 0, z or 0]

        self.M = self._build_matrix(T=T) @ self.M

    def rotate(self, R: np.ndarray = None, x: float = None, y: float = None, z: float = None):
        r"""
        Add a rotational component to transformation matrix.
        Arguments can be given as a rotation matrix R \in R^3x3 or by specifying the angle to rotate along the x, y or z axes

        Example usage:
            | ``Transform.rotate(get_rotmat(x=1, y=-1))``
            | ``Transform.rotate(x=1, y=-1)``

        .. seealso::

            :func:`get_rotmat`
            :func:`rotate`
        """
        if R is None:
            R = get_rotmat(x=x, y=y, z=z)

        self.M = self._build_matrix(R=R) @ self.M

    def scale(self, S: np.ndarray = None, x: float = None, y: float = None, z: float = None):
        """
        Add a scaling component to the transformation matrix.
        Arguments can be given as a container of x, y, z values.
        You can also specify x, y and z components separately

        Example usage:
            | ``Transform.scale([0, 0, 3])``
            | ``Transform.scale(z=3)``
        """
        if S is None:
            S = [x or 1, y or 1, z or 1]
        elif isinstance(S, (float, int)):
            S = [S, S, S]

        self.M = self._build_matrix(S=S) @ self.M

    def reflect(self, normal: np.ndarray = None):
        """
        Add a reflection across a plane given by a normal vector to the transformation matrix.
        The reflection is given as

        :math:`R = \\mathbb{I} - 2\\frac{nn^T}{n^Tn} \\in \\mathbb{R}^{3 \\times 3}`

        where :math:`n` is the normal vector of the plane to reflect along.

        Args:
            normal: the normal vector of the plane to reflect across. 
                If not given or ``None``, it will be set to one unit along the x-axis, i.e. a reflection along the yz-plane.

        References:
            https://en.wikipedia.org/wiki/Reflection_(mathematics)
        """
        if normal is None:
            normal = np.array([[1, 0, 0]])

        normal = np.atleast_2d(np.array(normal))

        # normalize the normal to be sure
        normal = normal.T / np.linalg.norm(normal)
        R = np.eye(3) - 2 * (normal @ normal.T) / (normal.T @ normal)
        self.M = self.M @ self._build_matrix(R=R)

    def _build_matrix(self, R: np.ndarray = None, T: np.ndarray = None, S: np.ndarray = None) -> np.ndarray:
        '''
        Build the transformation matrix for this object.
        '''
        # set the default matrix and vectors
        R = R if R is not None else get_rotmat()
        T = T if T is not None else np.array([0, 0, 0])
        S = S if S is not None else np.array([1, 1, 1])

        return np.array([[R[0, 0] * S[0], R[0, 1], R[0, 2], T[0]], [R[1, 0], R[1, 1] * S[1], R[1, 2], T[1]], [R[2, 0], R[2, 1], R[2, 2] * S[2], T[2]], [0, 0, 0, 1]])

    def get_rotmat(self):
        return self.M[:3, :3]

    def get_translation(self):
        return self.M[:3, 3]

    def to_vtkTransform(self):
        import vtk
        vtktrans = vtk.vtkTransform()
        vtktrans.PostMultiply()

        angles = rotmat_to_angles(self.get_rotmat())
        vtktrans.RotateX(angles[0] * 180 / np.pi)
        vtktrans.RotateY(angles[1] * 180 / np.pi)
        vtktrans.RotateZ(angles[2] * 180 / np.pi)

        vtktrans.Translate(self.get_translation())

        return vtktrans



class KabschTransform(Transform):
    """
    Use Kabsch-Umeyama algorithm to calculate the optimal transformation matrix :math:`T_{Kabsch}` that minimizes the 
    RMSD between two sets of coordinates :math:`X \\in \\mathbb{R}^{N \\times 3}` and :math:`Y \\in \\mathbb{R}^{N \\times 3}`, such that 
    
    :math:`\\text{arg}\\min_{T_{Kabsch}} \\text{RMSD}(T_{Kabsch}(X), Y)`

    It is numerically stable and works when the covariance matrix is singular.
    Both sets of points must be the same size for this algorithm to work.
    The coordinates are first centered onto their centroids before determining the optimal rotation matrix.

    Args:
        X: array containing the first set of coordinates. The Kabsch transformation matrix will be made such that applying it to ``X`` will yield ``Y``.
        Y: array containing the second set of coordinates. These coordinates is the target to transform to.

    .. warning::
        In principle, the Kabsch-Umeyama algorithm does not care about the dimensions of the coordinates, 
        however we will always assume 3D coordinates as that is our most common use-case. Further, the :class:`Transform` class also assumes 3D coordinates. 
        If you would like to make use of 2D or 1D Transforms we suggest you simply set the correct axes to zero.

    .. seealso::
        :class:`Transform`
            The main transformation class.

    Example:

        .. code-block:: python

            from tcutility import geometry
            import numpy as np
    
            # create two arrays that are the same
            X, Y = np.arange(5 * 3).reshape(5, 3), np.arange(5 * 3).reshape(5, 3)  

            # create a transformation matrix to change X
            Tx = geometry.Transform()
            Tx.rotate(x=1, y=1, z=1)
            Tx.translate(x=1, y=1, z=1)

            X = Tx(X)
            
            # get the Kabsch transformation matrix
            Tkabsch = geometry.KabschTransform(X, Y)
    
            # check if applying the transformation matrix to X yields Y
            assert np.isclose(Tkabsch(X), Y).all()


    References:
        | https://en.wikipedia.org/wiki/Orthogonal_Procrustes_problem
        | https://en.wikipedia.org/wiki/Kabsch_algorithm
    """
    def __init__(self, X: np.ndarray, Y: np.ndarray):
        # make sure arrays are 2d and the same size
        X, Y = np.atleast_2d(X), np.atleast_2d(Y)
        assert X.shape == Y.shape, f"Matrices X with shape {X.shape} and Y with shape {Y.shape} are not the same size"

        # center the coordinates
        centroid_x = np.mean(X, axis=0)
        centroid_y = np.mean(Y, axis=0)
        Xc = X - centroid_x
        Yc = Y - centroid_y

        # calculate covariance matrix H
        H = Xc.T @ Yc

        # first do single value decomposition on covariance matrix
        # this step ensures that the algorithm is numerically stable
        # and removes problems with singular covariance matrices
        U, _, V = scipy.linalg.svd(H)

        # get the sign of the determinant of V.T @ U.T
        sign = np.sign(np.linalg.det(V.T @ U.T))
        # then build the optimal rotation matrix
        d = np.diag([1, 1, sign])
        R = V.T @ d @ U.T

        # build the transformation:
        # We have that Y ~= (R @ (X - centroid_x).T).T + centroid_y
        # the normal order is to first translate X by -centroid_x
        # then rotate with R
        # finally translate by +centroid_y
        self.M = self._build_matrix()
        self.translate(-centroid_x)
        self.rotate(R)
        self.translate(centroid_y)


class MolTransform(Transform):
    '''
    A subclass of :class:`Transform` that is designed to generate transformation for a molecule.
    It adds, among others, methods for aligning atoms to specific vectors, planes, or setting the centroid of the molecule.
    The nice thing is that the class applies the transformations based only on the atom indices given by the user.

    Args:
        mol: the molecule that is used for the alignment.

    .. note::
        Indexing starts at 1 instead of 0.
    '''
    def __init__(self, mol: plams.Molecule):
        self.mol = mol
        super().__init__()

    def center(self, *indices):
        '''
        Center the molecule on given indices or by its centroid.

        Args:
            indices: the indices that are used to center the molecule. 
                If not given the centering will be done based on all atoms.
        '''
        tmol = self.apply(self.mol)
        if len(indices) == 0:
            indices = range(1, len(tmol) + 1)
        C = np.array([tmol.as_array()[i - 1] for i in indices])
        self.translate(-np.mean(C, axis=0))

    def align_to_vector(self, index1: int, index2: int, vector: Sequence[float] = None):
        '''
        Align the molecule such that a bond lays on a given vector.

        Args:
            index1: index of the first atom.
            index2: index of the second atom.
            vector: the vector to align the atoms to. If not given or `None` it defaults to `(1, 0, 0)`.
        '''
        # get the transformed mol
        tmol = self.apply(self.mol)
        # and coordinates
        C1, C2 = tmol.as_array()[index1 - 1], tmol.as_array()[index2 - 1]
        if vector is None:
            vector = [1, 0, 0]

        R = vector_align_rotmat(C1 - C2, vector)
        self.rotate(R)

    def align_to_plane(self, index1: int, index2: int, index3: int, vector: Sequence[float] = None):
        '''
        Align a molecule such that the normal of the plane defined by three atoms is aligned to a given vector.

        Args:
            index1: index of the first atom.
            index2: index of the second atom.
            index3: index of the third atom.
            vector: the vector to align the atoms to. If not given or `None` it defaults to (0, 1, 0).
        '''
        # get the transformed mol
        tmol = self.apply(self.mol)
        # and coordinates
        C1, C2, C3 = tmol.as_array()[index1 - 1], tmol.as_array()[index2 - 1], tmol.as_array()[index3 - 1]
        if vector is None:
            vector = [0, 1, 0]

        # calculate normal vector and align it to the given vector
        n = np.cross(C1 - C2, C3 - C2)
        R = vector_align_rotmat(n, vector)
        self.rotate(R)


def get_rotmat(x: float = None, y: float = None, z: float = None) -> np.ndarray:
    """
    Create a rotation matrix based on the Tait-Bryant sytem.
    In this system, x, y, and z are angles of rotation around
    the corresponding axes. This function uses the right-handed
    convention

    Args:
        x: Rotation around the x-axis in radians.
        y: Rotation around the y-axis in radians.
        z: Rotation around the z-axis in radians.

    Returns:
        the rotation matrix :math:`\\textbf{R} \\in \\mathbb{R}^{3 \\times 3}` with the specified axis rotations.


    .. seealso::

        :func:`apply_rotmat`
            For applying the rotation matrix to coordinates.

        :func:`rotate`
            For rotating coordinates directly, given Tait-Bryant angles.

        :meth:`Transform.rotate`
            The :class:`Transform` class allows you to also rotate.
    """
    # start with identity matrix
    R = np.eye(3)

    # apply rotation around each axis separately
    if x is not None:
        c = cos(x)
        s = sin(x)
        R = R @ np.array(([1, 0, 0], [0, c, -s], [0, s, c]))

    if y is not None:
        c = cos(y)
        s = sin(y)
        R = R @ np.array(([c, 0, s], [0, 1, 0], [-s, 0, c]))

    if z is not None:
        c = cos(z)
        s = sin(z)
        R = R @ np.array(([c, -s, 0], [s, c, 0], [0, 0, 1]))

    return R


def rotmat_to_angles(R: np.ndarray) -> Tuple[float]:
    thetax = atan2(R[2, 1], R[2, 2])
    thetay = atan2(-R[2, 0], sqrt(R[2, 1]**2 + R[2, 2]**2))
    thetaz = atan2(R[1, 0], R[0, 0])
    return thetax, thetay, thetaz


def apply_rotmat(coords: np.ndarray, R: np.ndarray) -> np.ndarray:
    """
    Apply a rotation matrix to a set of coordinates.

    Args:
        coords: the coordinates :math`\\in \\mathbb{R}^{n \\times 3}` to rotate.
        R: the rotation matrix to apply. 

    Returns:
        New coordinates :math`\\in \\mathbb{R}^{n \\times 3}` rotated using the given rotation matrix.

    .. seealso::

        :func:`get_rotmat`
            For creating a rotation matrix.

        :func:`rotate`
            For rotating coordinates directly, given Tait-Bryant angles.
    """
    coords = np.atleast_2d(coords)
    return (R @ coords.T).T.squeeze()


def rotate(coords: np.ndarray, x: float = None, y: float = None, z: float = None) -> np.ndarray:
    """
    Build and apply a rotation matrix to a set of coordinates.

    Args:
        coords: the coordinates :math`\\in \\mathbb{R}^{n \\times 3}` to rotate.
        x: Rotation around the x-axis in radians.
        y: Rotation around the y-axis in radians.
        z: Rotation around the z-axis in radians.

    .. seealso::
        :func:`get_rotmat`
            For creating a rotation matrix.
    """
    return apply_rotmat(coords, get_rotmat(x, y, z))


def vector_align_rotmat(a: np.ndarray, b: np.ndarray) -> np.ndarray:
    """
    Calculate a rotation matrix that aligns vector a onto vector b.

    Args:
        a: vector that is to be aligned.
        b: vector that is the target of the alignment.

    Returns:
        Rotation matrix R, such that ``geometry.apply_rotmat(a, R) == b``.
    """
    # normalize the vectors first
    a = np.array(a) / np.linalg.norm(a)
    b = np.array(b) / np.linalg.norm(b)

    c = a @ b
    if c == 1:
        # if a == b we simply return the identity matrix
        return np.eye(3)
    if c == -1:
        # when a == -b we cannot directly calculate R, as 1/(1+c) is undefined
        # instead, we first create a random rotation matrix and apply it to a
        # to get a new vector aprime. We then align aprime to b, which is possible since aprime != -b
        # to get the final rotation matrix we simply multiply the random and alignment rotation matrices
        Rrand = get_rotmat(np.pi / 3, np.pi / 3, np.pi / 3)
        return vector_align_rotmat(apply_rotmat(a, Rrand), b) @ Rrand

    v = np.cross(a, b)
    skew = np.array([[0, -v[2], v[1]], [v[2], 0, -v[0]], [-v[1], v[0], 0]])
    R = np.eye(3) + skew + skew @ skew / (1 + c)
    return R


def RMSD(X: np.ndarray, Y: np.ndarray, axis: Union[int, None] = None, use_kabsch: bool = True, include_mirror: bool = False) -> float:
    r"""
    Calculate Root Mean Squared Deviations between two sets of points ``X`` and ``Y``.
    By default Kabsch' algorithm is used to align the sets of points prior to calculating the RMSD.
    Optionally the axis can be given to calculate the RMSD along different axes.

    RMSD is given as

    :math:`\text{RMSD}(X, Y) = \frac{1}{N}\sqrt{\sum_i^N (X_i - Y_i)^2}`

    when using the Kabsch algorithm to align the two sets of coordinates we first obtain the :class:`KabschTransform` :math:`T_{Kabsch}` and then

    :math:`\text{RMSD}(X, Y) = \frac{1}{N}\sqrt{\sum_i^N (T_{Kabsch}(X_i) - Y_i)^2}`

    Args:
        X: the first set of coordinates to compare. It must have the same dimensions as ``Y``.
        Y: the second set of coordinates to compare. It must have the same dimensions as ``X``.
        axis: axis to compare. Defaults to ``None``. 
        use_kabsch: whether to use Kabsch' algorithm to align ``X`` and ``Y`` before calculating the RMSD. Defaults to ``True``.
        include_mirror: return the lowest value between the RMSD of the supplied coordinates and also the RMSD of mirrored X with Y.
            This will only be done if ``use_kabsch == True``.

    Returns:
        RMSD in the units of X and Y. If ``axis`` is set to an integer this function will return a vector of RMSD's along that axis.

    .. note::
        It is generally recommended to enable the use of the Kabsch-Umeyama algorithm prior to calculating the RMSD. 
        This will ensure you get the lowest possible RMSD for you sets of coordinates.

    .. seealso::
        :class:`KabschTransform`
    """

    X = np.array(X)
    Y = np.array(Y)

    assert X.shape == Y.shape

    # apply Kabsch transform
    if use_kabsch:
        Tkabsch = KabschTransform(X, Y)
        Xprime = Tkabsch(X)

    rmsd = np.sqrt(np.sum((Xprime - Y) ** 2, axis=axis) / X.shape[0])

    # if we include the mirror image we have to apply a reflection to the coordinates
    # and then recalculate the kabsch transform in the mirror coordinates
    # then we calculate the new RMSD and take the smaller of the new and old RMSD
    if include_mirror and use_kabsch:
        Tmirror = Transform()
        Tmirror.reflect()
        Tkabsch_mirror = KabschTransform(Tmirror(X), Y)
        Xprime = Tkabsch_mirror(Tmirror(X))
        rmsd_mirrored = np.sqrt(np.sum((Xprime - Y) ** 2, axis=axis) / X.shape[0])
        rmsd = min(rmsd, rmsd_mirrored)

    return rmsd


def random_points_on_sphere(shape: Tuple[int], radius: float = 1) -> np.ndarray:
    """
    Generate random points on a sphere with a specified radius.

    Args:
        shape: The shape of the resulting points, generally shape[0] coordinates with shape[1] dimensions
        radius: The radius of the sphere to generate the points on.

    Returns:
        Array of coordinates on a sphere.
    """
    x = np.random.randn(*shape)
    x = x / np.linalg.norm(x, axis=1, keepdims=True) * radius
    return x


def random_points_in_anular_sphere(shape: Tuple[int], min_radius: float = 0, max_radius: float = 1):
    """
    Generate random points in an sphere or anular sphere with specified radii. 
    An anular sphere is a hollow sphere of a certain thickness.

    Args:
        shape: The shape of the resulting points, generally shape[0] coordinates with shape[1] dimensions
        min_radius: The lowest radius of the sphere to generate the points in.
        max_radius: The largest radius of the sphere to generate the points in.

    Returns:
        Array of coordinates on a sphere.
    """
    random_radii = np.random.rand(shape[0]) * (max_radius - min_radius) + min_radius
    return random_points_on_sphere(shape, random_radii)


def random_points_on_spheroid(coordinates: np.ndarray, Nsamples: int = 1, margin: float = 0):
    """
    Generate random points on a spheroid generated by a set of coordinates.

    Args:
        coordinates: The (n x dim) set of coordinates that is used to generate the minimum-volume spheroid.
        Nsamples: The number of samples to return.
        margin: the spacing between the sampling spheroid and the minimum-volume spheroid.

    Returns:
        Array of coordinates on a spheroid.
    """
    # for this to work we should first get the centroid of our molecule
    centroid = np.mean(coordinates, axis=0)
    # and get the centered coordiantes
    Xc = coordinates - centroid

    # we then do a singular-value decomposition to obtain
    # the three principle components (Vh) with their eigenvalues (s)
    _, s, Vh = scipy.linalg.svd(Xc)

    # then compute a transformation matrix for generating the correct spheroid
    transform = Transform()
    transform.translate(centroid)
    transform.rotate((np.diag(s/2 + margin) @ Vh).T)

    # to sample the spheroid we generate points on a 
    # sphere and transform them to our spheroid
    p = random_points_on_sphere((Nsamples, Xc.shape[1]))
    return transform(p)


def parameter(coordinates, *indices, pyramidal=False):
    '''
    Return geometry information about a set of coordinates given indices.
    '''
    assert 1 <= len(indices) <= 4, "Number of indices must be between 1, 2, 3 or 4"

    coordinates = np.array(coordinates)
    selected_coords = [coordinates[i] for i in indices]

    if len(indices) == 1:
        return selected_coords[0]

    if len(indices) == 2:
        return np.linalg.norm(selected_coords[0] - selected_coords[1])

    if len(indices) == 3:
        a = selected_coords[0] - selected_coords[1]
        b = selected_coords[2] - selected_coords[1]
        a = a / np.linalg.norm(a)
        b = b / np.linalg.norm(b)

        return np.arccos(a @ b) / np.pi * 180

    if len(indices) == 4 and not pyramidal:
        a = selected_coords[0] - selected_coords[1]
        b = selected_coords[2] - selected_coords[1]

        u = selected_coords[1] - selected_coords[2]
        v = selected_coords[3] - selected_coords[2]

        n1, n2 = np.cross(a, b), np.cross(u, v)

        n1 = n1 / np.linalg.norm(n1)
        n2 = n2 / np.linalg.norm(n2)

        return np.arccos(n1 @ n2) / np.pi * 180


    if len(indices) == 4 and pyramidal:
        ang1 = parameter(coordinates, indices[1], indices[0], indices[2])
        ang2 = parameter(coordinates, indices[2], indices[0], indices[3])
        ang3 = parameter(coordinates, indices[3], indices[0], indices[1])

        return 360 - ang1 - ang2 - ang3



import json
import pathlib as pl
from tcutility import data


available_basis_sets = { 
'ADF': [
	'SZ',
	'DZ',
	'DZP',
	'TZP',
	'TZ2P',
	'QZ4P',
	'mTZ2P',
	'AUG/ASZ',
	'AUG/ADZ',
	'AUG/ADZP',
	'AUG/ATZP',
	'AUG/ATZ2P',
	'Corr/TZ3P',
	'Corr/QZ6P',
	'Corr/ATZ3P',
	'Corr/AQZ6P',
	'ET/ET-pVQZ',
	'ET/ET-QZ3P',
	'ET/ET-QZ3P-1DIFFUSE',
	'ET/ET-QZ3P-2DIFFUSE',
	'ET/ET-QZ3P-3DIFFUSE',
	'TZ2P-J',
	'QZ4P-J',
	'POLTDDFT/DZ',
	'POLTDDFT/DZP',
	'POLTDDFT/TZP',
],
'BAND': {},
'ORCA': {},
}


# read data
data_dir = pl.Path(__file__).parents[0] / "_atom_data_info"

with open(data_dir / "norbs.json") as inp:
	_number_of_orbitals = json.loads(inp.read())


def number_of_orbitals(element, basis_set):
	'''
	Get the number of atomic orbitals for a certain element and basis-set.

	Args:
		element: the element for which to get the number of AOs.
		basis_set: the basis-set for which to get the number of AOs.

	.. warning::
		This function currently only works for the following basis-sets: 
		[``SZ``, ``DZ``, ``DZP``, ``TZP``, ``TZ2P``, ``TZ2P-J``, ``mTZ2P``, ``QZ4P``, ``QZ4P-J``, ``jcpl``].
		It also only works for no-frozen-core calculations with ``NOSYM`` symmetry.
	'''
	symbol = data.atom.symbol(element)
	return _number_of_orbitals[basis_set][symbol]


def number_of_virtuals(element, basis_set):
	'''
	Get the number of virtual atomic orbitals for a certain element and basis-set.
	The number of virtuals is equal to the total number of AOs minus half the number of electrons in the atom.

	Args:
		element: the element for which to get the number of AOs.
		basis_set: the basis-set for which to get the number of AOs.

	.. warning::
		This function currently only works for the following basis-sets: 
		[``SZ``, ``DZ``, ``DZP``, ``TZP``, ``TZ2P``, ``TZ2P-J``, ``mTZ2P``, ``QZ4P``, ``QZ4P-J``, ``jcpl``].
		It also only works for no-frozen-core calculations with ``NOSYM`` symmetry.
	'''
	num = data.atom.atom_number(element)
	symbol = data.atom.symbol(element)
	return _number_of_orbitals[basis_set][symbol] - num/2


if __name__ == '__main__':
	print(number_of_orbitals('Pd', 'DZ'))
	print(number_of_virtuals('Pd', 'DZ'))



from tcutility import results
import os

j = os.path.join


def test_read_dft() -> None:
    res = results.read(j(os.path.split(__file__)[0], "fixtures", "DFT_EDA"))
    assert isinstance(res, results.Result)


def test_dft_engine() -> None:
    res = results.read(j(os.path.split(__file__)[0], "fixtures", "DFT_EDA"))
    assert res.engine == "adf"


def test_LOT() -> None:
    # LOT = level of theory
    res = results.read(j(os.path.split(__file__)[0], "fixtures", "DFT_EDA"))
    assert res.level.summary == "OLYP/TZ2P"


def test_LOT2() -> None:
    res = results.read(j(os.path.split(__file__)[0], "fixtures", "level_of_theory", "MP2_SOS"))
    assert res.level.summary == "MP2-SOS/TZ2P"


def test_LOT3() -> None:
    res = results.read(j(os.path.split(__file__)[0], "fixtures", "level_of_theory", "PBE0_MBD_rsSC"))
    assert res.level.summary == "PBE0-MBD@rsSC/TZ2P"


def test_LOT4() -> None:
    res = results.read(j(os.path.split(__file__)[0], "fixtures", "level_of_theory", "M06_2X"))
    assert res.level.summary == "M06-2X/QZ4P"


def test_sections() -> None:
    res = results.read(j(os.path.split(__file__)[0], "fixtures", "DFT_EDA"))
    assert all(section in res for section in ["adf", "engine", "ams_version", "is_multijob", "molecule", "status", "timing"])


def test_failed_reading() -> None:
    res = results.read('not/a/real/calculation')
    assert res.engine == 'unknown'


def test_failed_reading2() -> None:
    res = results.read('not/a/real/calculation')
    assert res.status.code == 'U'


def test_failed_reading3() -> None:
    res = results.read('not/a/real/calculation')
    assert res.status.fatal is True


if __name__ == '__main__':
    import pytest
    pytest.main()



from tcutility import molecule
import os


j = os.path.join

root_dir = j(os.path.split(__file__)[0], 'molecules')


def get(name):
    p = j(root_dir, f'{name.removesuffix(".xyz")}.xyz')
    return molecule.load(p)


def get_molecules(tags=None):
    for f in os.listdir(root_dir):
        if not f.endswith('.xyz'):
            continue
            
        mol = get(f)
        if tags is None or any(tag in mol.flags.tags for tag in tags):
            yield mol


if __name__ == '__main__':
    for mol in get_molecules():
        print(mol, mol.flags)



from typing import TypeVar
import numpy.typing as npt
import numpy as np

DType = TypeVar("DType", bound=np.generic)
Array1D = npt.NDArray[DType]
Array2D = npt.NDArray[DType]
Array3D = npt.NDArray[DType]



from scm import plams
from tcutility.data import molecules
from tcutility.job.generic import Job
from tcutility import log
import os


j = os.path.join


class CRESTJob(Job):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.crest_path = 'crest'
        self.xtb_path = 'xtb'
        self._charge = 0
        self._spinpol = 0
        self._temp = 400
        self._mdlen = 'x1'

    def _setup_job(self):
        self.add_postamble('mkdir rotamers')
        self.add_postamble('mkdir conformers')
        self.add_postamble(f'split -d -l {len(self._molecule.atoms) + 2} -a 5 crest_conformers.xyz conformers/')
        self.add_postamble(f'split -d -l {len(self._molecule.atoms) + 2} -a 5 crest_rotamers.xyz rotamers/')

        self.add_postamble('for file in conformers/* rotamers/*')
        self.add_postamble('do')
        self.add_postamble('    mv "$file" "$file.xyz"')
        self.add_postamble('done')

        os.makedirs(self.workdir, exist_ok=True)

        self._molecule.write(j(self.workdir, 'coords.xyz'))

        options = [
            'coords.xyz',
            f'-xnam "{self.xtb_path}"',
            '--noreftopo',
            f'-c {self._charge}',
            f'-u {self._spinpol}',
            f'-tnmd {self._temp}',
            f'-mdlen {self._mdlen}',
        ]

        options = ' '.join(options)

        with open(self.runfile_path, 'w+') as runf:
            runf.write('#!/bin/sh\n\n')  # the shebang is not written by default by ADF
            runf.write('\n'.join(self._preambles) + '\n\n')
            runf.write(f'{self.crest_path} {options}\n')
            runf.write('\n'.join(self._postambles))

        return True

    def spin_polarization(self, val: int):
        '''
        Set the spin-polarization of the system.
        '''
        self._spinpol = val

    def multiplicity(self, val: int):
        '''
        Set the multiplicity of the system. If the value is not one the calculation will also be unrestricted.
        We use the following values:
        
        1) singlet
        2) doublet
        3) triplet
        4) ...

        The multiplicity is equal to 2*S+1 for spin-polarization of S.
        '''
        self._spinpol = (val - 1)//2

    def charge(self, val: int):
        '''
        Set the charge of the system.
        '''
        self._charge = val

    def md_temperature(self, val: float):
        '''
        Set the temperature of the molecular dynamics steps. Defaults to 400K.
        '''
        self._temp = val

    def md_length(self, val: float):
        '''
        Set the length of the molecular dynamics steps. The default length will be multiplied by this value, e.g. the default value is 1.
        '''
        self._mdlen = f'x{val}'

    @property
    def best_conformer_path(self):
        return j(self.workdir, 'crest_best.xyz')

    @property
    def conformer_directory(self):
        return j(self.workdir, 'conformers')

    @property
    def rotamer_directory(self):
        return j(self.workdir, 'rotamers')

    def get_conformer_xyz(self, number: int = None):
        '''
        Return paths to conformer xyz files for this job.

        Args:
            number: the number of files to return, defaults to 10. If the directory already exists, for example if the job was already run, we will return up to `number` files.
        '''
        if os.path.exists(self.conformer_directory):
            return [j(self.conformer_directory, file) for i, file in enumerate(sorted(os.listdir(self.conformer_directory)))]

        for i in range(number or 10):
            yield j(self.conformer_directory, f'{str(i).zfill(5)}.xyz')

    def get_rotamer_xyz(self, number: int = None):
        '''
        Return paths to rotamer xyz files for this job.

        Args:
            number: the number of files to return, defaults to 10. If the directory already exists, for example if the job was already run, we will return up to `number` files.
        '''
        if os.path.exists(self.rotamer_directory):
            return [j(self.rotamer_directory, file) for i, file in enumerate(os.listdir(self.rotamer_directory))]

        for i in range(number or 10):
            yield j(self.rotamer_directory, f'{str(i).zfill(5)}.xyz')


class QCGJob(CRESTJob):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.crest_path = 'crest'
        self._nsolv = 10
        self._fix_solute = False
        self._ensemble_generation_mode = 'NCI-MTD'
        self._solvent = None
        self._nofix = True

    def solvent(self, mol, nsolv=None):
        if self._nsolv:
            self._nsolv = nsolv

        if isinstance(mol, plams.Molecule):
            self._solvent = mol

        elif isinstance(mol, str) and os.path.exists(mol):
            self._solvent = plams.Molecule(mol)

        elif isinstance(mol, str):
            # here we can get the molecule name using tcutility.data.molecules
            self._solvent = molecules.get(mol)
            # we can check if the alpb solvent name is defined in the xyz file
            alpb = self._solvent.flags.get('alpb', None)
            if alpb:
                self.alpb(alpb)

        elif isinstance(mol, list) and isinstance(mol[0], plams.Atom):
            self._solvent = plams.Molecule()
            [self._solvent.add_atom(atom) for atom in mol]

        elif isinstance(mol, plams.Atom):
            self._solvent = plams.Molecule()
            self._solvent.add_atom(mol)

    def nsolv(self, nsolv):
        self._nsolv = nsolv

    def alpb(self, solvent):
        self._alpb = solvent

    def ensemble_mode(self, mode):
        self._ensemble_generation_mode = mode

    def nofix(self, enable=True):
        self._nofix = enable

    def _setup_job(self):
        self.add_postamble('mkdir ensemble/ensemble')
        self.add_postamble(f'split -d -l {len(self._molecule.atoms) + len(self._solvent.atoms) * self._nsolv + 2} -a 5 ensemble/final_ensemble.xyz ensemble/ensemble')

        self.add_postamble('for file in ensemble/ensemble/*')
        self.add_postamble('do')
        self.add_postamble('    mv "$file" "$file.xyz"')
        self.add_postamble('done')

        if not self._solvent:
            log.error(f'Did not provide a solvent molecule for this job. Call the {self.__class__.__name__}.solvent method to add one.')

        os.makedirs(self.workdir, exist_ok=True)

        self._molecule.write(j(self.workdir, 'coords.xyz'))
        self._solvent.write(j(self.workdir, 'solvent.xyz'))

        ensemble_mode_option = {
            'NCI-MTD': '--ncimtd',
            'MD': '--md',
            'MTD': '--mtd',
        }.get(self._ensemble_generation_mode, self._ensemble_generation_mode)

        options = [
            'coords.xyz',
            f'-xnam "{self.xtb_path}"',
            '--qcg solvent.xyz',
            '--ensemble',
            f'--nsolv {self._nsolv}',
            f'--chrg {self._charge}',
            f'--uhf {self._spinpol}',
            f'--tnmd {self._temp}',
            f'--mdlen {50 * float(self._mdlen[1:])}',
            ensemble_mode_option,
        ]

        if self._alpb:
            options.append(f'--alpb {self._alpb}')

        if self._nofix:
            options.append('--nofix')

        options = ' '.join(options)

        with open(self.runfile_path, 'w+') as runf:
            runf.write('#!/bin/sh\n\n')
            runf.write('\n'.join(self._preambles) + '\n\n')
            runf.write(f'{self.crest_path} {options}\n')
            runf.write('\n'.join(self._postambles))

        return True

    @property
    def ensemble_directory(self):
        return j(self.workdir, 'ensemble', 'ensemble')

    def get_ensemble_xyz(self, number: int = None):
        '''
        Return paths to conformer xyz files for this job.

        Args:
            number: the number of files to return, defaults to 10. If the directory already exists, for example if the job was already run, we will return up to `number` files.
        '''
        if os.path.exists(self.ensemble_directory):
            return [j(self.ensemble_directory, file) for i, file in enumerate(sorted(os.listdir(self.ensemble_directory)))]

        for i in range(number or 10):
            yield j(self.ensemble_directory, f'{str(i).zfill(5)}.xyz')

    @property
    def best_ensemble_path(self):
        return j(self.workdir, 'ensemble', 'ensemble', 'crest_best.xyz')


if __name__ == '__main__':
    # with CRESTJob() as job:
    #     job.rundir = 'tmp/SN2'
    #     job.name = 'CREST'
    #     job.molecule('../../../test/fixtures/xyz/transitionstate_radical_addition.xyz')
    #     job.sbatch(p='tc', ntasks_per_node=32)

    with QCGJob(test_mode=True) as job:
        job.rundir = 'calculations/Ammonia'
        job.name = 'QCG'
        job.molecule('ammonia.xyz')
        job.solvent('water', 10)
        print(job._solvent)
        job.sbatch(p='tc', n=32)

    # for i in range(40):
    #     with CRESTJob(test_mode=False, overwrite=True) as job:
    #         job.rundir = 'tmp/SN2'
    #         job.name = 'CREST'
    #         job.molecule('../../../test/fixtures/xyz/transitionstate_radical_addition.xyz')
    #         job.sbatch(p='tc', ntasks_per_node=32)



'''
Module used for obtaining information about exchange-correlation functionals.
For example, it can be useful to obtain
'''
import os
from tcutility import results, cache
import re


j = os.path.join


@cache.cache
def get(functional_name: str) -> results.Result:
    '''
    Return information about a given functional.

    Args:
        functional_name: the name of the functional. It should exist in the :func:`get_available_functionals` keys.

    Return:
        A :class:`Result <tcutility.results.result.Result>` object containing information about the functional if it exists. Else it will return ``None``.

    .. seealso::
        :func:`get_available_functionals` for an overview of the information returned.
    '''
    info = functionals.get(functional_name, None)
    if info is None:
        info = functionals.get(functional_name_from_path_safe_name(functional_name), None)

    if info is None:
        raise KeyError(f'Could not find info for functional {functional_name}.')

    return info


def functional_name_from_path_safe_name(path_safe_name: str) -> results.Result:
    '''
    Return information about a given functional given its path-safe name.
    This can be useful when you want to know the functional from a path name.

    Return:
        A :class:`Result <tcutility.results.result.Result>` object containing information about the functional if it exists. Else it will return ``None``.

    .. seealso::
        :func:`get_available_functionals` for an overview of the information returned.
    '''
    for functional, functional_info in functionals.items():
        if path_safe_name.lower() == functional_info.path_safe_name.lower():
            return functional
        elif path_safe_name.replace('-', '').lower() == functional_info.path_safe_name.replace('-', '').lower():
            return functional



def get_available_functionals():
    '''
    Function that returns a dictionary of all available XC-functionals.

    Returns:
        : A :class:`Result <tcutility.results.result.Result>` object containing information about all available XC-functionals.
            The functional names are stored as the keys and the functional information is stored as the values.
            The values contain the following information:

                - ``name`` **(str)** - the name of the functional.
                - ``path_safe_name`` **(str)** - the name of the functional made suitable for file paths.
                    This name is the same as the normal name, but without parentheses. Asterisks are replaced with lower-case ``s``.
                - ``name_no_disp`` **(str)** - the name of functional without the dispersion correction.
                - ``category`` **(str)** - the category the functional belongs to.
                - ``dispersion`` **(str)** - the dispersion correction part of the functional name.
                - ``dispersion_name`` **(str)** - the name of the dispersion correction as it would be written in ADF.
                - ``includes_disp`` **(bool)** - whether the functional already includes a dispersion correction.
                - ``use_libxc`` **(bool)** - whether the functional is from the LibXC library.
                - ``available_in_adf`` **(bool)** - whether the functional is available in ADF.
                - ``available_in_band`` **(bool)** - whether the functional is available in BAND.
                - ``available_in_orca`` **(bool)** - whether the functional is available in ORCA.
                - ``adf_settings`` **(:class:`Result <tcutility.results.result.Result>`)** - the settings that are used to select the functional in the ADF input.
                - ``name_latex`` **(str)** - the name of the functional formatted to be used with LaTeX renderers.
                - ``name_html`` **(str)** - the name of the functional formatted to be used with HTML renderers.
                - ``dois`` **(List[str])** - a list of relevant dois for this functional.
    '''
    def set_dispersion(func):
        disp_map = {
            '-D4': 'GRIMME4',
            '-D3(BJ)': 'GRIMME3 BJDAMP',
            '-D3BJ': 'GRIMME3 BJDAMP',
            '-D3': 'GRIMME3',
            '-dDsC': 'dDsC',
            '-dUFF': 'UFF',
            '-MBD': 'MBD',
            '-MBD@rsSC': 'MBD',
            '-D': 'DEFAULT'
        }

        # set some default values for useful parameters
        func.name_no_disp = func.name
        func.dispersion = None
        func.dispersion_name = None

        # go through every dispersion correction and check if we are using it
        for disp_suffix, disp_name in disp_map.items():
            if func.name.endswith(disp_suffix):
                # dispersion will be the suffix without the -
                func.dispersion = disp_suffix[1:]
                func.dispersion_name = disp_name

                # check if the functional already includes the dispersion correction
                if func.includes_disp:
                    break

                # else we set the name of the functional without dispersion for later
                func.name_no_disp = func.name[:-len(disp_suffix)]
                # get the dispersion settings for ADF. Try to get custom values if they were provided.
                func.adf_settings.XC.Dispersion = func.disp_params or disp_name
                break

    def set_functional(func):
        # set the functional settings for ADF
        # first go through some special functionals that require special settings
        if func.name_no_disp == 'BMK':
            func.adf_settings.XC.LibXC = 'HYB_MGGA_X_BMK GGA_C_BMK'
            return

        if func.name in ['LCY-BLYP', 'LCY-BP86', 'LCY-PBE']:
            func.adf_settings.XC.GGA = func.name.split('-')[1]
            func.adf_settings.XC.RANGESEP = ''
            func.adf_settings.XC.xcfun = ''
            return

        if func.name in ['CAMY-B3LYP']:
            func.adf_settings.XC.Hybrid = 'CAMY-B3LYP'
            func.adf_settings.XC.RANGESEP = ''
            func.adf_settings.XC.xcfun = ''
            return

        if func.name == 'GGA:SSB-D':
            func.adf_settings.XC.GGA = 'SSB-D'
            return

        if func.name == 'MetaGGA:SSB-D':
            func.adf_settings.XC.MetaGGA = 'SSB-D'
            return

        if func.name_no_disp == 'HartreeFock':
            func.adf_settings.XC.HartreeFock = ''
            return

        if func.name == 'MP2':
            func.adf_settings.XC.MP2 = ''
            return

        if func.name in ['SOS-MP2', 'SCS-MP2']:
            func.adf_settings.XC.MP2 = ''
            func.adf_settings.XC.EmpiricalScaling = func.name[:-4]
            return

        # the normal functionals are defined based on their category, or selected from libxc
        if func.use_libxc:
            func.adf_settings.XC.LibXC = func.name_no_disp
        else:
            func.adf_settings.XC[func.category] = func.name_no_disp

    # gather all data about available functionals
    functionals = results.Result()  # store all info in this dict

    with open(j(os.path.split(__file__)[0], 'available_functionals.txt')) as file:
        lines = file.readlines()

    # read the references first
    # references are given as
    # [refname] doi
    # and in-line
    # - xcname {options} [refname1][refname2]...
    dois = {}
    for line in lines:
        if not line.startswith('['):
            continue

        line = line.split('#')[0].strip()

        ref, doi = line.strip().split()
        ref = ref[1:-1]
        dois[ref] = doi

    for line in lines:
        # there can be empty lines
        if not line.strip():
            continue

        # and comment lines
        if line.startswith('#'):
            continue

        # we don't read the references here
        if line.startswith('['):
            continue

        # functional names are given starting with -
        # category names without -
        if not line.startswith('- '):
            curr_category = line.strip()
            continue

        line = line.split('#')[0].strip()

        # store data about the func in a dict
        func = results.Result()
        func.category = curr_category
        # separate the functional name from the line
        functional_name = line[2:].split('!')[0].split(',')[0].split('[')[0].strip()
        func.name = functional_name
        func.name_latex = functional_name
        func.name_html = functional_name
        func.path_safe_name = functional_name.replace(')', '').replace('(', '').replace('*', 's').replace(' ', '-')

        if functional_name.startswith('WB'):
            func.name_latex = func.name_latex.replace('WB', r'$\omega$B')
            func.name_html = func.name_html.replace('WB', '&omega;B')

        if 'r2SCAN' in functional_name:
            func.name_latex = func.name_latex.replace('r2SCAN', r'r$^2$SCAN')
            func.name_html = func.name_html.replace('r2SCAN', 'r<sup>2</sup>SCAN')

        if 'and' in functional_name:
            func.name_latex = func.name_latex.replace('and', '&')
            func.name_html = func.name_html.replace('and', '&amp;')

        if '*' in functional_name:
            func.name_latex = func.name_latex.replace('*', r'$^*$')
            func.name_html = func.name_html.replace('*', '<sup>*</sup>')

        if 'B2PIPLYP' in functional_name:
            func.name_latex = func.name_latex.replace('B2PIPLYP', r'B2$\pi$PLYP')
            func.name_html = func.name_html.replace('B2PIPLYP', 'B2&pi;PLYP')

        # check if custom params were given for dispersion
        if 'GRIMME' in line:
            # func.disp_params = line.split('!')[0].split(',')[1].strip().strip("'")
            func.disp_params = re.findall(r"'(.*)'", line)

        func.use_libxc = '!libxc' in line
        func.includes_disp = '!includesdisp' in line
        func.available_in_adf = '!noadf' not in line
        func.available_in_band = '!band' in line
        func.available_in_orca = '!orca' in line

        # get references
        refs = re.findall(r'\[([a-zA-Z0-9]+)\]', line)
        func.dois = [dois[ref] for ref in refs]

        set_dispersion(func)
        set_functional(func)

        functionals[functional_name] = func

    return functionals

functionals = get_available_functionals()

categories = []
for functional in functionals:
    if get(functional).category not in categories:
        categories.append(get(functional).category)



"""
This module provides basic and general information about calculations done using AMS given a calculation directory.
This includes information about the engine used (ADF, DFTB, BAND, ...), general information such as timings, files, status of the calculation, etc.
This information is used in further analysis programs.

"""

from typing import Union

from . import result

Result = result.Result

import os  # noqa: E402
import pathlib as pl  # noqa: E402

from .. import slurm  # noqa: E402
from . import adf, ams, cache, dftb, orca, xtb, crest  # noqa: E402


def get_info(calc_dir: str):
    try:
        return ams.get_ams_info(calc_dir)
    except:  # noqa
        pass

    try:
        return orca.get_info(calc_dir)
    except:  # noqa
        pass

    try:
        return xtb.get_info(calc_dir)
    except:  # noqa
        pass

    try:
        return crest.get_info(calc_dir)
    except:  # noqa
        pass


    res = Result()

    # if we cannot correctly read the info, we return some generic result object
    # before that, we check the job status by checking slurm
    if slurm.workdir_info(os.path.abspath(calc_dir)):
        res.engine = "unknown"

        state = slurm.workdir_info(os.path.abspath(calc_dir)).statuscode
        state_name = {"CG": "COMPLETING", "CF": "CONFIGURING", "PD": "PENDING", "R": "RUNNING"}.get(state, "UNKNOWN")

        res.status.fatal = False
        res.status.name = state_name
        res.status.code = state
        res.status.reasons = []
    else:
        res.engine = "unknown"
        res.status.fatal = True
        res.status.name = "UNKNOWN"
        res.status.code = "U"
        res.status.reasons = []

        # give a little more specific information about the error
        if not os.path.exists(calc_dir):
            res.status.reasons.append(f"Could not find folder {calc_dir}")
        elif not os.path.isdir(calc_dir):
            res.status.reasons.append(f"Path {calc_dir} is not a directory")
        else:
            res.status.reasons.append(f"Could not read calculation in {calc_dir}")

    return res


def read(calc_dir: Union[str, pl.Path]) -> Result:
    """Master function for reading data from calculations. It reads general information as well as engine-specific information.

    Args:
        calc_dir: path pointing to the working directory for the desired calculation

    Returns:
        dictionary containing information about the calculation
    """
    calc_dir = str(calc_dir) if isinstance(calc_dir, pl.Path) else calc_dir

    ret = Result()
    ret.update(get_info(calc_dir))
    if ret.engine == "adf":
        try:
            ret.adf = adf.get_calc_settings(ret)
        except:  # noqa
            ret.adf = None

        try:
            ret.properties = adf.get_properties(ret)
        except:  # noqa
            ret.properties = None

        try:
            ret.level = adf.get_level_of_theory(ret)
        except:  # noqa
            ret.level = None

    elif ret.engine == "dftb":
        ret.dftb = dftb.get_calc_settings(ret)
        ret.properties = dftb.get_properties(ret)
    elif ret.engine == "xtb":
        # ret.xtb = xtb.get_calc_settings(ret)
        ret.properties = xtb.get_properties(ret)

    elif ret.engine == "orca":
        try:
            ret.orca = orca.get_calc_settings(ret)
        except:  # noqa
            ret.orca = None

        try:
            ret.properties = orca.get_properties(ret)
        except:  # noqa
            ret.properties = None

    # unload cached KFReaders associated with this calc_dir
    to_delete = [key for key in cache._cache if key.startswith(os.path.abspath(calc_dir))]
    [cache.unload(key) for key in to_delete]
    return ret


if __name__ == "__main__":
    res = read("/Users/yumanhordijk/Library/CloudStorage/OneDrive-VrijeUniversiteitAmsterdam/RadicalAdditionBenchmark/data/abinitio/P_C2H2_NH2/OPT_pVTZ")
    print(res.molecule)



import numpy as np
import warnings
from tcutility import results
from scm import plams


def avg_relative_bond_length_delta(base: plams.Molecule, pos: plams.Molecule, neg: plams.Molecule, atom1: int, atom2: int) -> float:
    """Function to calculate relative atom distance change in vibrational mode

    Args:
        base: plams.Molecule object containing the coordinates before applying vibration
        pos: plams.Molecule object with the vibrational mode added
        neg: plams.Molecule object with the vibrational mode subtracted
        atom1: label for the first atom
        atom2: label for the second atom

    Returns:
        Average relative difference from the baseline distance between selected atoms in this vibrational mode (as percentage).
    """
    basedist = base[atom1].distance_to(base[atom2])
    x = abs((pos[atom1].distance_to(pos[atom2]) / basedist) - 1)  # distances between positive and negative vibration are normalized with base distance
    y = abs((neg[atom1].distance_to(neg[atom2]) / basedist) - 1)  # variances are obtained and averaged
    return (x + y) / 2


def determine_ts_reactioncoordinate(data: results.Result, mode_index: int = 0, bond_tolerance: float = 1.28, min_delta_dist: float = 0.0) -> np.ndarray:
    """Function to retrieve reaction coordinate from a given transitionstate, using the first imaginary frequency.

    Args:
        data: TCutility.results.Result object containing calculation data
        mode_index: vibrational mode index to analyze
        bond_tolerance: parameter for plams.Molecule.guess_bonds() function
        min_delta_dist: minimum relative bond length change before qualifying as active atom. If 0, all bond changes are counted

    Returns:
        Array containing all the obtained reaction coordinates. Reaction coordinate format is [active_atom1, active_atom2, sign], using Distance reactioncoordinate
        Symmetry elements are ignored, by convention the atom labels are increasing (atom1 < atom2)
    """

    assert "modes" in data.properties.vibrations, "Vibrational data is required, but was not present in .rkf file"

    outputmol = data.molecule.output

    base = np.array(outputmol).reshape(-1, 3)
    tsimode = np.array(data.properties.vibrations.modes[mode_index]).reshape(-1, 3)

    posvib = outputmol.copy()
    posvib.from_array(base + tsimode)
    posvib.guess_bonds(dmax=bond_tolerance)
    negvib = outputmol.copy()
    negvib.from_array(base - tsimode)
    negvib.guess_bonds(dmax=bond_tolerance)

    pairs = np.where(posvib.bond_matrix() - negvib.bond_matrix())
    rc = np.array(
        [
            [a + 1, b + 1, np.sign(posvib.bond_matrix()[a][b] - negvib.bond_matrix()[a][b])]
            for a, b in np.c_[pairs[0], pairs[1]]  # bond matrices and pairs matrix are 0-indexed unlike the plams.Molecule labels, hence a+1 and b+1 are needed
            if a < b and avg_relative_bond_length_delta(outputmol, posvib, negvib, a + 1, b + 1) > min_delta_dist
        ],
        dtype=int,
    )
    return rc


def validate_transitionstate(calc_dir: str, rcatoms: list = None, analyze_modes: int = 1, **kwargs) -> bool:
    """Function to determine whether a transition state calculation yielded the expected transition state. Checks the reaction coordinates provided by the user (or in the .rkf file User Input section) and compares
        this against the reaction coordinates found in the imaginary modes of the transitionstate. If the transitionstate has multiple imaginary frequencies, it is possible to check multiple modes for the expected
        reaction coordinate.

    Args:
        calc_dir: path pointing to the desired calculation.
        rcatoms: list or array containing expected reaction coordinates, to check against the transition state. If not defined, it is obtained from the ams.rkf user input.
                  Only uses 'Distance' reaction coordinate. Format should be [atomlabel1, atomlabel2, (optional) sign]
        analyze_modes: Number of imaginary modes to analyze, default only the first mode. Modes are ordered lowest frequency first. If 0 or negative value is provided, analyze all modes with imaginary frequency.
        **kwargs: keyword arguments for use in :func:`determine_ts_reactioncoordinate`.

    Returns:
        Boolean value, True if the found transition state reaction coordinates contain the expected reaction coordinates, False otherwise.
        If multiple modes are analyzed, returns True if at least one mode contains the expected reaction coordinates.
    """

    data = results.read(calc_dir)
    assert "modes" in data.properties.vibrations, "Vibrational data is required, but was not present in .rkf file"

    nimag = sum(1 for x in data.properties.vibrations.frequencies if x < 0)
    if nimag == 0:
        return False  # no imaginary modes found in transitionstate

    if analyze_modes > 0:  # if positive value is provided, check that many imaginary vibrational modes. Otherwise all imaginary ones are analyzed
        nimag = min(nimag, analyze_modes)  # limit to the actual number of vibrational modes

    if isinstance(rcatoms, type(None)):
        assert "reactioncoordinate" in data.input.transitionstatesearch, "Reaction coordinate is a required input, but was neither provided nor present in the .rkf file"
        rcatoms = np.array([[int(float(y)) for y in x.split()[1:]] for x in data.input.transitionstatesearch.reactioncoordinate if x.split()[0] == "Distance"])
        assert len(rcatoms) > 0, "Reaction coordinate data was present in .rkf file, but no reaction coordinate using Distance was provided"

    # cast the reaction coordinate as 2d array. If reaction coordinate sign is provided, it must be provided for all reaction coordinates.
    # the np.atleast_2d() function will cause a deprecation warning if the dimensions mismatch, this is raised as a value error instead
    with warnings.catch_warnings(record=True) as w:
        rcatoms = np.atleast_2d(rcatoms)
        if len(w) > 0:
            raise ValueError(w[-1].message)

    assert np.all([len(x) == 2 or len(x) == 3 for x in rcatoms]), "Invalid format of reaction coordinate. Reaction coordinate format must be [label1, label2, (optional) sign]"

    ret = []

    for idx in range(nimag):
        result = determine_ts_reactioncoordinate(data, mode_index=idx, **kwargs)
        if len(result) == 0:
            ret.append(False)
            continue

        # sort for consistency
        for index, [a, b, *c] in enumerate(rcatoms):
            if c:
                c = np.sign(c)
            if a > b:
                a, b = b, a
            rcatoms[index] = [a, b, *c]
        rcatoms = rcatoms[rcatoms[:, 1].argsort()]
        rcatoms = rcatoms[rcatoms[:, 0].argsort()]
        result = result[result[:, 1].argsort()]
        result = result[result[:, 0].argsort()]

        # (optional) check for internal consistency if the reaction coordinate sign is provided
        # only the first reaction coordinate sign is arbitrary, check remaining coordinates for consistency
        if len(rcatoms[0]) > 2:
            foundmatch = False
            for idx, rc in enumerate(result):
                if rc[0] == rcatoms[0][0] and rc[1] == rcatoms[0][1]:
                    result[:, 2] = result[:, 2] if np.sign(result[idx][2]) == rcatoms[0][2] else -result[:, 2]
                    foundmatch = True
                    break
            if not foundmatch:  # at least one element of rcatoms is not present in result
                ret.append(False)
                continue
        else:
            result = result[:, :-1]

        ret.append(set(str(x) for x in rcatoms).issubset(set(str(y) for y in result)))

    return any(ret)



from tcutility.job.generic import Job
from tcutility import spell_check
import os
from typing import Union


j = os.path.join


class XTBJob(Job):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.xtb_path = 'xtb'
        self._options = ['coords.xyz']
        self._scan = {}

    def _setup_job(self):
        os.makedirs(self.workdir, exist_ok=True)

        self._molecule.write(j(self.workdir, 'coords.xyz'))

        if self._scan:
            with open(j(self.workdir, 'scan.inp'), 'w+') as scaninp:
                for key, value in self._scan.items():
                    scaninp.write(f'${key}\n')
                    for line in value:
                        scaninp.write(f'    {line}\n')
                scaninp.write('$end\n')

        options = ' '.join(self._options)
        with open(self.runfile_path, 'w+') as runf:
            runf.write('#!/bin/sh\n\n')  # the shebang is not written by default by ADF
            runf.write('\n'.join(self._preambles) + '\n\n')
            runf.write(f'{self.xtb_path} {options}\n')
            runf.write('\n'.join(self._postambles))

        return True

    def model(self, method: Union[str, int]):
        '''
        Set the method used by XTB. This includes GFN0-xTB, GFN1-xTB, GFN2-xTB and GFNFF.

        Args:
            method: the method to use. Can be specified by its full name, e.g. 'GFN2-xTB', is the same as 'GFN2' or simply 2.
        '''
        if isinstance(method, int):
            if not 0 >= method >= 2:
                raise ValueError(f'GFN{method}-xTB does not exist. Please choose one of GFN[0, 1, 2]-xTB.')
            self._options.append(f'--gfn {method}')
            return

        if method.lower() in ['gfn0', 'gfn0-xtb']:
            self._options.append('--gfn 0')
            return

        if method.lower() in ['gfn1', 'gfn1-xtb']:
            self._options.append('--gfn 1')
            return

        if method.lower() in ['gfn2', 'gfn2-xtb']:
            self._options.append('--gfn 2')
            return

        if method.lower() in ['gfnff', 'ff']:
            self._options.append('--gfnff')
            return

        raise ValueError(f'Did not recognize the method {method} for XTB.')
    
    def solvent(self, name: str = None, model: str = 'alpb'):
        '''
        Model solvation using the ALPB or GBSA model.

        Args:
            name: the name of the solvent you want to use. Must be ``None``, ``Acetone``, ``Acetonitrile``, ``CHCl3``, ``CS2``, ``DMSO``, ``Ether``, ``H2O``, ``Methanol``, ``THF`` or ``Toluene``.
            grid_size: the size of the grid used to construct the solvent accessible surface. Must be ``230``, ``974``, ``2030`` or ``5810``.
        '''
        spell_check.check(model, ['alpb', 'gbsa'], ignore_case=True)
        self._options.append(f'--{model} {name}')

    def spin_polarization(self, val: int):
        '''
        Set the spin-polarization of the system.
        '''
        self._options.append(f'-u {val}')

    def multiplicity(self, val: int):
        '''
        Set the multiplicity of the system. If the value is not one the calculation will also be unrestricted.
        We use the following values:
        
        1) singlet
        2) doublet
        3) triplet
        4) ...

        The multiplicity is equal to 2*S+1 for spin-polarization of S.
        '''
        self._options.append(f'-u {(val - 1)//2}')

    def charge(self, val: int):
        '''
        Set the charge of the system.
        '''
        self._options.append(f'-c {val}')

    def vibrations(self, enable: bool = True):
        self._options.append('--hess')

    def optimization(self, quality: str = 'Normal', calculate_hess: bool = True):
        '''
        Do a geometry optimization and calculate the normal modes of the input structure.
        
        Args:
            quality: the convergence criteria of the optimization. 
                See https://xtb-docs.readthedocs.io/en/latest/optimization.html#optimization-levels.
            calculate_hess: whether to calculate the Hessian and do a normal mode analysis after optimization.
        '''
        spell_check.check(quality, ['crude', 'sloppy', 'loose', 'lax', 'normal', 'tight', 'vtight', 'extreme'], ignore_case=True)

        if calculate_hess:
            self._options.append(f'--ohess {quality}')
        else:
            self._options.append(f'--opt {quality}')

    def PESScan(self, distances: list = [], angles: list = [], dihedrals: list = [], npoints: int = 20, quality: str = 'Normal', mode: str = 'concerted'):
        '''
        Set the task of the job to potential energy surface scan (PESScan).

        Args:
            distances: sequence of tuples or lists containing ``[atom_index1, atom_index2, start, end]``. 
                Atom indices start at 1. Distances are given in |angstrom|.
            angles: sequence of tuples or lists containing ``[atom_index1, atom_index2, atom_index3, start, end]``. 
                Atom indices start at 1. Angles are given in degrees
            dihedrals: sequence of tuples or lists containing ``[atom_index1, atom_index2, atom_index3, atom_index4, start, end]``. 
                Atom indices start at 1. Angles are given in degrees
            npoints: the number of PES points to optimize.

        .. note::
            Currently we only support generating settings for 1-dimensional PESScans. 
            We will add support for N-dimensional PESScans later.
        '''
        self.optimization(quality=quality, calculate_hess=False)
        self._options.append('--input scan.inp')

        self._scan.setdefault('scan', [f'mode={mode}'])
        # self._scan.setdefault('opt', [f'maxcycles=50'])
        for i, d in enumerate(distances):
            self._scan['scan'].append(f'distance: {d[0]},{d[1]},{d[2]}; {d[2]}, {d[3]}, {npoints}')
        for i, a in enumerate(angles, start=i):
            self._scan['scan'].append(f'angle: {a[0]},{a[1]},{a[2]},{a[3]}; {a[3]}, {a[4]}, {npoints}')
        for i, a in enumerate(dihedrals, start=i):
            self._scan['scan'].append(f'dihedral: {a[0]},{a[1]},{a[2]},{a[3]},{a[4]}; {a[4]}, {a[5]}, {npoints}')

    @property
    def output_mol_path(self):
        return f'{self.workdir}/xtbopt.xyz'




from typing import Union
from scm import plams


def parse_molecule(molecule: plams.Molecule) -> str:
    """
    Analyse a molecule and return the molstring describing its parts. Each part will then be separated by a ``+`` sign in the new string.

    Args:
        molecule: ``plams.Molecule`` object to be parsed.

    Returns:
        A string that contains each part of the molecule separated by a ``+`` sign, for use in :func:`molecule` function for further formatting.
    """
    # to separate a molecule we need to have bonds
    molecule.guess_bonds()
    parts = []
    # go through each part of the molecule
    for part in molecule.separate():
        # get a dictionary of counts for each element
        form = part.get_formula(True)
        # add all elements with their number, but only if the number is larger than 1.
        # this prevents the creation of strings containing e.g. C1H3Cl2
        parts.append("".join([sym + (str(num) if num > 1 else "") for sym, num in form.items()]))

    return " + ".join(parts)


def molecule(molecule: Union[str, plams.Molecule], mode: str = "unicode") -> str:
    """
    Parse and return a string containing a molecular formula that will show up properly in LaTeX, HTML or unicode.

    Args:
        molecule: ``plams.Molecule`` object or a string that contains the molecular formula to be parsed. 
            It can be either single molecule or a reaction. Molecules should be separated by ``+`` or ``->``.
        mode: the formatter to convert the string to. Should be ``unicode``, ``html``, ``latex``, ``pyplot``.

    Returns:
        A string that is formatted to be rendered nicely in either HTML or LaTeX. 
        In the returned strings any numbers will be subscripted and ``+``, ``-``, ``*`` and ``•`` will be superscripted. 
        For ``latex`` and ``pyplot`` modes we apply ``\\mathrm`` to letters.

    Examples:
        >>> molecule('C9H18NO*')
        'C₉H₁₈NO•'

        >>> molecule('C2H2 + CH3* -> C2H2CH3', mode='html')
        'C<sub>2</sub>H<sub>2</sub> + CH<sub>3</sub><sup>•</sup> -> C<sub>2</sub>H<sub>2</sub>CH3'

    .. seealso::
        The :func:`parse_molecule` function is used to convert ``plams.Molecule`` objects to a molecular formula.

    """
    # to take care of plus-signs used to denote reactions we have to first split
    # the molstring into its parts.
    if isinstance(molecule, plams.Molecule):
        molstring = parse_molecule(molecule)
    else:
        molstring = molecule

    molstring = molstring.replace("*", "•")
    for part in molstring.split():
        # if part is only a plus-sign we skip this part. This is only true when the plus-sign
        # is used to denote a reaction
        if part in ["+", "->"]:
            continue

        # parse the part
        partret = part
        # numbers should be subscript
        for num in "0123456789":
            if mode in ["latex", "pyplot"]:
                partret = partret.replace(num, f"_{num}")
            if mode == "html":
                partret = partret.replace(num, f"<sub>{num}</sub>")
            if mode == "unicode":
                partret = partret.replace(num, "₀₁₂₃₄₅₆₇₈₉"[int(num)])

        partret_ = partret
        partret = ''
        for char in partret_:
            if char.isalpha() and mode in ["latex", "pyplot"]:
                partret += f"\mathrm{{{char}}}"
            else:
                partret += char

        # signs should be superscript
        for sign in "+-•":
            # negative charges should be denoted by em dash and not a normal dash
            if mode in ["latex", "pyplot"]:
                partret = partret.replace(sign, f'^{sign.replace("-", "—")}')
            if mode == "html":
                partret = partret.replace(sign, f'<sup>{sign.replace("-", "—")}</sup>')
            if mode == "unicode":
                partret = partret.replace(sign, "⁺⁻•"["+-•".index(sign)])
        # replace the part in the original string
        molstring = molstring.replace(part, partret)

    if mode == 'pyplot':
        return fr"${molstring}$"

    return molstring


if __name__ == "__main__":
    # print(molecule("F- + CH3Cl", "html"))
    # mol = plams.Molecule(r"D:\Users\Yuman\Desktop\PhD\TCutility\test\fixtures\chloromethane_sn2_ts\ts sn2.results\output.xyz")
    # print(molecule(mol))

    print(molecule('NMe2*', mode='latex'))


import pathlib as pl
from typing import Dict, List, Union

from scm import plams

from tcutility import ensure_list
from tcutility.results import result
from tcutility.data import atom


def number_of_electrons(mol: plams.Molecule, charge: int = 0) -> int:
    """
    The number of electrons in a molecule.

    Args:
        mol: the molecule to count the number of electrons from.
        charge: the charge of the molecule.

    Returns:
        The sum of the atomic numbers in the molecule minus the charge of the molecule.
    """
    nel = 0
    for at in mol:
        nel += atom.atom_number(at.symbol)
    return nel - charge


def _parse_str(s: str):
    # checks if string should be an int, float, bool or string

    # sanitization first
    # empty s returns None
    # if s is not a str return it
    if s == "":
        return None
    if not isinstance(s, str):
        return s

    if "," in s:
        return [_parse_str(part.strip()) for part in s.split(",")]

    # to parse the string we use try/except method
    try:
        return int(s)
    except ValueError:
        pass

    try:
        return float(s)
    except ValueError:
        pass

    # bool type casting always works, so we have to specifically check for correct strings
    if s in ["True", "False"]:
        return bool(s)

    return s


def load(path) -> plams.Molecule:
    """
    Load a molecule from a given xyz file path.
    The xyz file is structured as follows:

    .. code-block::

        [int]
        Comment line
        [str] [float] [float] [float] atom_tag1 atom_tag2 atom_key1=...
        [str] [float] [float] [float] atom_tag1 atom_tag2 atom_key1=...
        [str] [float] [float] [float]

        mol_tag1
        mol_tag2
        mol_key1=...
        mol_key2 = ...


    The xyz file is parsed and returned as a :class:`plams.Molecule <plams.mol.molecule.Molecule>` object.
    Flags and tags are given as ``mol.flags`` and ``mol.flags.tags`` respectively.
    Similarly for the atoms, the flags and tags are given as ``mol.atoms[i].flags`` and ``mol.atoms[i].flags.tags``
    """

    def parse_flags(args):
        ret = result.Result()
        ret.tags = set()
        for arg in args:
            # flags are given as key=value pairs
            # tags are given as loose keys
            if "=" in arg:
                key, value = arg.split("=")
                ret[key.strip()] = _parse_str(value.strip())
            else:
                ret.tags.add(_parse_str(arg.strip()))
        return ret

    with open(path) as f:
        lines = [line.strip() for line in f.readlines()]

    mol = plams.Molecule()

    natoms = int(lines[0])
    mol.comment = lines[1]
    # not all lines in the file will define atoms. Lines after line natoms+2 will be used for flags
    atom_lines = lines[2 : natoms + 2]
    for line in atom_lines:
        # parse every atom first
        symbol, x, y, z, *args = line.split()
        at = plams.Atom(symbol=symbol, coords=(float(x), float(y), float(z)))
        at.flags = parse_flags(args)
        mol.add_atom(at)

    # after the atoms we parse the flags for the molecule
    flag_lines = lines[natoms + 2 :]
    flag_lines = [line.strip() for line in flag_lines if line.strip()]
    mol.flags = parse_flags(flag_lines)

    return mol


def from_string(mol: str) -> plams.Molecule:
    """
    Load a molecule from a string. Currently only supports simple XYZ-files, 
    e.g. not extended XYZ-files with flags.

    Args:
        mol: string containing the molecule to parse.
            This function only reads the element, x, y and z coordinates on each line.

    Returns:
        A new molecule object with the elements and coordinates from the input.
    """
    mol = plams.Molecule()
    for line in mol.splitlines():
        parts = [_parse_str(part) for part in line.split()]
        if len(parts) < 4:
            continue

        # check if first part is the symbol of the molecule
        if not isinstance(parts[0], str) and len(parts[0]) > 2:
            continue

        # check if
        if not all(isinstance(part, (float, int)) for part in parts[1:4]):
            continue

        atom = plams.Atom(symbol=parts[0], coords=parts[1:4])
        mol.add_atom(atom)

    return mol


def guess_fragments(mol: plams.Molecule) -> Dict[str, plams.Molecule]:
    """
    Guess fragments based on data from the xyz file. Two methods are currently supported, see the tabs below.
    We also support reading of charges and spin-polarizations for the fragments.
    They should be given as ``charge_{fragment_name}`` and ``spinpol_{fragment_name}`` respectively.


    .. tabs::

        .. group-tab:: First method

            .. code-block::

                8

                N       0.00000000       0.00000000      -0.81474153
                B      -0.00000000      -0.00000000       0.83567034
                H       0.47608351      -0.82460084      -1.14410295
                H       0.47608351       0.82460084      -1.14410295
                H      -0.95216703       0.00000000      -1.14410295
                H      -0.58149793       1.00718395       1.13712667
                H      -0.58149793      -1.00718395       1.13712667
                H       1.16299585      -0.00000000       1.13712667

                frag_Donor = 1, 3-5
                frag_Acceptor = 2, 6-8
                charge_Donor = -1
                spinpol_Acceptor = 2

            In this case, fragment atom indices must be provided below the coordinates.
            The fragment name must be prefixed with ``frag_``. Indices can be given as integers or as ranges using ``-``.

        .. group-tab:: Second method

            .. code-block::

                8

                N       0.00000000       0.00000000      -0.81474153 frag=Donor
                B      -0.00000000      -0.00000000       0.83567034 frag=Acceptor
                H       0.47608351      -0.82460084      -1.14410295 frag=Donor
                H       0.47608351       0.82460084      -1.14410295 frag=Donor
                H      -0.95216703       0.00000000      -1.14410295 frag=Donor
                H      -0.58149793       1.00718395       1.13712667 frag=Acceptor
                H      -0.58149793      -1.00718395       1.13712667 frag=Acceptor
                H       1.16299585      -0.00000000       1.13712667 frag=Acceptor

                charge_Donor = -1
                spinpol_Acceptor = 2

            In this case, fragment atoms are marked with the `frag` flag which gives the name of the fragment the atom belongs to.

    Args:
        mol: the molecule that is to be split into fragments. It should have defined either method shown above.
             If it does not define these methods this function returns ``None``.

    Returns:
        A dictionary containing fragment names as keys and :class:`plams.Molecule <scm.plams.mol.molecule.Molecule>` objects as values.
        Atoms that were not included by either method will be placed in the molecule object with key ``None``.

    """

    # first method, check if the fragments are defined as molecule flags
    fragment_flags = [flag for flag in mol.flags if flag.startswith("frag_")]

    if len(fragment_flags) > 0:
        # we split here to get of the frag_ prefix
        fragment_mols = {frag.split("_", 1)[1]: plams.Molecule() for frag in fragment_flags}

        for frag in fragment_flags:
            frag_name = frag.split("_", 1)[1]
            indices = []
            index_line = ensure_list(mol.flags[frag])
            for indx in index_line:
                if isinstance(indx, int):
                    indices.append(indx)
                elif isinstance(indx, str) and "-" in indx:
                    indices.extend(range(int(indx.split("-")[0]), int(indx.split("-")[1]) + 1))
                else:
                    raise ValueError(f"Fragment index {indx} could not be parsed.")

            [fragment_mols[frag_name].add_atom(mol[i]) for i in indices]
            fragment_mols[frag_name].flags = {"tags": set()}
            if f"charge_{frag_name}" in mol.flags:
                fragment_mols[frag_name].flags["charge"] = mol.flags[f"charge_{frag_name}"]
            if f"spinpol_{frag_name}" in mol.flags:
                fragment_mols[frag_name].flags["spinpol"] = mol.flags[f"spinpol_{frag_name}"]

        return result.Result(fragment_mols)

    # second method, check if the atoms have a frag= flag defined
    fragment_names = set(at.flags.get("frag") for at in mol)
    if len(fragment_names) > 0:
        fragment_mols = {name: plams.Molecule() for name in fragment_names}
        for at in mol:
            # get the fragment the atom belongs to and add it to the list
            fragment_mols[at.flags.get("frag")].add_atom(at)

        for frag in fragment_names:
            fragment_mols[frag].flags = {"tags": set()}
            if f"charge_{frag}" in mol.flags:
                fragment_mols[frag].flags["charge"] = mol.flags[f"charge_{frag}"]
            if f"spinpol_{frag}" in mol.flags:
                fragment_mols[frag].flags["spinpol"] = mol.flags[f"spinpol_{frag}"]

        return result.Result(fragment_mols)


# =============================================================================
# Writing Functions for Molecule objects ======================================
# =============================================================================


def _xyz_format(mol: plams.Molecule, include_n_atoms: bool = True) -> str:
    """Returns a string representation of a molecule in the xyz format, e.g.:

    C      0.00000000      0.00000000      0.00000000
    H      1.00000000      0.00000000      0.00000000
    H      0.00000000      1.00000000      0.00000000
    H      0.00000000      0.00000000      1.00000000

    ...
    """
    n_atoms = len(mol.atoms)
    if include_n_atoms:
        return f"{n_atoms}\n" + "\n".join([f"{at.symbol:6s}{at.x:16.8f}{at.y:16.8f}{at.z:16.8f}" for at in mol.atoms])

    return "\n".join([f"{at.symbol:6s}{at.x:16.8f}{at.y:16.8f}{at.z:16.8f}" for at in mol.atoms])


def _amv_format(mol: plams.Molecule, step: int, energy: Union[float, None] = None, name: Union[float, str] = None) -> str:
    """Returns a string representation of a molecule in the amv format, e.g.:

    Geometry 1, Energy: -0.5 Ha
    C      0.00000000      0.00000000      0.00000000
    H      1.00000000      0.00000000      0.00000000
    H      0.00000000      1.00000000      0.00000000
    H      0.00000000      0.00000000      1.00000000
    ...

    If no energy is provided, the energy is not included in the string representation"""
    header = f"Geometry {step}"
    header += f", Name: {name}" if name is not None else ""
    header += f", Energy: {energy} Ha" if energy is not None else ""

    return header + "\n" + "\n".join([f"{at.symbol:6s}{at.x:16.8f}{at.y:16.8f}{at.z:16.8f}" for at in mol.atoms])


def write_mol_to_xyz_file(out_file: Union[str, pl.Path], mols: Union[List[plams.Molecule], plams.Molecule], include_n_atoms: bool = False) -> None:
    """Writes a list of molecules to a file in xyz format."""
    mols = mols if isinstance(mols, list) else [mols]
    out_file = pl.Path(f"{out_file}.xyz")

    [mol.delete_all_bonds() for mol in mols]
    write_string = "\n\n".join([_xyz_format(mol, include_n_atoms) for mol in mols])
    out_file.write_text(write_string)

    return None


def write_mol_to_amv_file(out_file: Union[str, pl.Path], mols: Union[List[plams.Molecule], plams.Molecule], energies: Union[List[float], None], mol_names: Union[List[str], None] = None) -> None:
    """Writes a list of molecules to a file in amv format."""
    out_file = pl.Path(out_file)

    if out_file.suffix != ".amv":
        out_file = out_file.with_suffix(".amv")

    mols = mols if isinstance(mols, list) else [mols]
    energies = energies if energies is not None else [0.0 for _ in mols]
    names = mol_names if mol_names is not None else [f"Molecule {i}" for i in range(1, len(mols) + 1)]

    [mol.delete_all_bonds() for mol in mols]
    write_string = "\n\n".join([_amv_format(mol, step, energy, name) for step, (mol, energy, name) in enumerate(zip(mols, energies, names), 1)])
    out_file.write_text(write_string)

    return None


def save(mol: plams.Molecule, path: str, comment: str = None):
    """Save a molecule in a custom xyz file format.
    Molecule and atom flags can be provided as the "flags" parameter of the object (mol.flags and atom.flags).
    """
    comment = comment or mol.comment if hasattr(mol, "comment") else ""
    with open(path, "w+") as f:
        f.write(f"{len(mol.atoms)}\n{comment}\n")
        for at in mol.atoms:
            flags_str = ""
            flags = at.flags if hasattr(at, "flags") else {}
            for key, value in flags.items():
                if key == "tags":
                    if len(value) > 0:
                        flags_str += " ".join([str(x) for x in value]) + " "
                    else:
                        continue
                else:
                    flags_str += f"{key}={value} "

            f.write(f"{at.symbol}\t{at.coords[0]: .10f}\t{at.coords[1]: .10f}\t{at.coords[2]: .10f}\t{flags_str}\n")

        f.write("\n")

        flags = mol.flags if hasattr(mol, "flags") else {}
        for key, value in flags.items():
            if key == "tags":
                f.write("\n".join([str(x) for x in value]) + "\n")
            else:
                f.write(f"{key} = {value}\n")


if __name__ == "__main__":
    # mol = load(r"D:\Users\Yuman\Desktop\PhD\ychem\calculations2\5ef3a511141a993d83552515df6bf882a7560dd806f88df4e2c2887b4d2a9595\transitionstate\input_mol.xyz")
    mol = load("../../examples/job/NH3BH3.xyz")
    frags = guess_fragments(mol)
    print(frags)



"""Module containing errors to distinguish between tcutility-specific errors and general python errors from other packages / scripts."""

from tcutility.log import loadbar

some_list = {"a": 1, "b": 2, "c": 3}
loadbar(some_list)


class TCError(Exception):
    """Base class for all errors in the tcutility package."""

    pass


class TCJobError(TCError):
    """An error that occurs when a job fails to run properly."""

    def __init__(self, job_class: str, message: str):
        self.job_class = job_class
        self.message = message
        super().__init__(f"Error in job class {job_class}: {message}")


class TCMoleculeError(TCError):
    """An error that occurs when a molecule is not in a valid state."""

    pass


# -----------------
# ADF-related errors
# -----------------


class TCCompDetailsError(TCError):
    """An error that occurs when the computation details are not in a valid state. It expects a section such as a "Functional" or "Basis set" and a message."""

    def __init__(self, section: str, message: str):
        self.section = section
        self.message = message
        super().__init__(f"Error in {section}: {message}")



import os
import subprocess as sp
from typing import List, Union

from scm import plams

from tcutility import ensure_list, log, results, slurm, spell_check
from tcutility.errors import TCMoleculeError
from tcutility.job.generic import Job

j = os.path.join


class ORCAJob(Job):
    def __init__(self, use_tmpdir=False, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.settings = results.Result()
        self.settings.main = set()
        self._charge = 0
        self._multiplicity = 1
        self._ghosts = []
        self._method = None
        self.memory = None
        self.processes = None
        self.orca_path = None
        self.use_tmpdir = use_tmpdir

        self.single_point()

    def main(self, val: Union[str, List[str]]):
        """
        Add main options for this ORCA calculation, they will be added to the input prepended with exclamation marks.

        Args:
            val: the main options to add. This can be a string or a list of strings with the main options.
        """
        # we want to split a string such as 'CC-pVTZ Opt CCSD(T)' into loose parts and add them separately
        # this should always return a list
        if isinstance(val, str):
            val = val.split()
        # add each
        [self.settings.main.add(key) for key in val]

    def remove_main(self, val: Union[str, List[str]]):
        if isinstance(val, str):
            val = val.split()

        lower_main = {key.casefold(): key for key in self.settings.main}
        for v in val:
            if v.casefold() in lower_main:
                self.settings.main.discard(lower_main[v.casefold()])

    def __remove_task(self):
        [self.remove_main(task) for task in ["sp", "opt", "optts", "neb-ts"]]

    def method(self, method):
        spell_check.check(method, ["HF", "MP2", "CCSD", "CCSD(T)", "CCSDT"])
        self.settings.main.add(method)
        self._method = method

    def reference(self, ref):
        spell_check.check(ref, ["UNO", "UHF", "UKS", "RHF", "RKS", "ROHF", "ROKS"])
        self.settings.main.add(ref)
        self._method = ref

    def QRO(self, enable=True):
        self.settings.MDCI.UseQROs = enable

    def basis_set(self, value):
        self.settings.main.add(value)

    def single_point(self):
        self.__remove_task()
        self.settings.main.add("SP")

    def transition_state(self):
        self.__remove_task()
        self.vibrations()
        self.settings.main.add("OptTS")

    def optimization(self):
        self.__remove_task()
        self.vibrations()
        self.settings.main.add("Opt")

    def vibrations(self, enable=True, numerical=False):
        self.remove_main("NumFreq")
        self.remove_main("Freq")
        if not enable:
            return
        if numerical:
            self.settings.main.add("NumFreq")
        else:
            self.settings.main.add("Freq")

    def charge(self, val):
        self._charge = val

    def spin_polarization(self, val):
        self._multiplicity = val + 1

    def multiplicity(self, val):
        self._multiplicity = val

    def ghost_atoms(self, indices: Union[int, List[int]]):
        self._ghosts.extend(ensure_list(indices))

    def get_memory_usage(self):
        mem = self.memory or self._sbatch.mem or None

        ntasks = self.processes
        if ntasks is None:
            if self._sbatch.n:
                ntasks = self._sbatch.n
            if self._sbatch.ntasks:
                ntasks = self._sbatch.ntasks
            if self._sbatch.ntasks_per_node:
                ntasks = self._sbatch.ntasks_per_node * self._sbatch.get("N", 1) * self._sbatch.get("nodes", 1)

        return mem, ntasks

    def molecule(self, mol: Union[str, plams.Molecule, plams.Atom, List[plams.Atom]], natoms: int = None):
        """
        Add a molecule to this calculation in various formats.

        Args:
            mol: the molecule to read, can be a path (str). If the path exists already we read it. If it does not exist yet, it will be read in later. mol can also be a plams.Molecule object or a single or a list of plams.Atom objects.
            natoms: If the molecule is supplied as a path you should also give the number of atoms.
        """
        super().molecule(mol)
        self.natoms = natoms

    def get_input(self):
        # set the correct memory usage and processes
        mem, ntasks = self.get_memory_usage()
        if ntasks and mem:
            if self._molecule is not None:
                natoms = len(self._molecule) - len(self._ghosts)
            else:
                if not hasattr(self, "natoms") or self.natoms is None:
                    raise TCMoleculeError("You set the molecule as a path and did not supply the number of atoms.")
                natoms = self.natoms

            ntasks = min(ntasks, (natoms - 1) * 3)
            self.settings.PAL.nprocs = ntasks
            self.settings.maxcore = int(mem / ntasks * 0.75)
        else:
            log.warn("MaxCore and nprocs not specified. Please use SBATCH settings or set job.processes and job.memory.")

        ret = ""
        for key in self.settings.main:
            ret += f"!{key}\n"
        ret += "\n"

        for option, block in self.settings.items():
            if option == "main":
                continue

            if isinstance(block, results.Result):
                ret += f"%{option}\n"

                for key, val in block.items():
                    ret += f"    {key} {val}\n"

                ret += "END\n\n"
            else:
                ret += f"%{option} {block}\n"

        ret += "\n"

        if self._molecule_path:
            ret += f"* xyzfile {self._charge} {self._multiplicity} {os.path.abspath(self._molecule_path)}\n"

        else:
            ret += f"* xyz {self._charge} {self._multiplicity}\n"
            for i, atom in enumerate(self._molecule, start=1):
                if i in self._ghosts:
                    ret += f"    {atom.symbol:2}: {atom.x: >13f} {atom.y: >13f} {atom.z: >13f}\n"
                else:
                    ret += f"    {atom.symbol:3} {atom.x: >13f} {atom.y: >13f} {atom.z: >13f}\n"
            ret += "*\n"

        return ret

    def _setup_job(self):
        try:
            if self.orca_path is None and not self.test_mode:
                self.orca_path = sp.check_output(["which", "orca"]).decode().strip()
        except sp.CalledProcessError:
            log.warn(f'Could not find the orca path. Set the {self.__class__.__name__}.orca_path attribute to add it. Now setting it to "$(which orca)", make sure the orca executable is findable.')
            self.orca_path = "$(which orca)"

        if not self._molecule and not self._molecule_path:
            log.error(f"You did not supply a molecule for this job. Call the {self.__class__.__name__}.molecule method to add one.")
            return

        os.makedirs(self.workdir, exist_ok=True)
        with open(self.inputfile_path, "w+") as inp:
            inp.write(self.get_input())

        with open(self.runfile_path, "w+") as runf:
            runf.write("#!/bin/sh\n\n")
            runf.write("\n".join(self._preambles) + "\n\n")

            # when using temporary directories for SLURM we need to do some extra setup
            # this is mainly moving the calculation directory to the TMPDIR location
            # and after the jobs is finished we copy back the results and remove the TMPDIR
            if self.use_tmpdir and slurm.has_slurm():
                runf.write('export TMPDIR="$TMPDIR/$SLURM_JOB_ID"\n')
                runf.write("mkdir -p $TMPDIR\n")
                runf.write("cd $TMPDIR\n")
                runf.write(f"cp {self.inputfile_path} $TMPDIR\n")

                runf.write(f"{self.orca_path} $TMPDIR/{self.name}.in\n")

                runf.write(f"cp $TMPDIR/* {self.workdir}\n")
                runf.write("rm -rf $TMPDIR\n")

            else:
                runf.write(f"{self.orca_path} {self.inputfile_path}\n")

            runf.write("\n".join(self._postambles))

        return True

    @property
    def output_mol_path(self):
        """
        The default file path for output molecules when running ADF calculations. It will not be created for singlepoint calculations.
        """
        return j(self.workdir, "OPT.xyz")


if __name__ == "__main__":
    job = ORCAJob()
    job.main("OPT cc-pVTZ")
    job.remove_main("OPT OPTTS NEB")



from tcutility import results
import docx
from htmldocx import HtmlToDocx


class SI:
    def __init__(self, path: str):
        """
        Class for creating supporting information (SI) files in Microsoft Word.

        Args:
                path: the location of the Word file. Does not have to have a file-extension.
                append_mode: whether to append to or overwrite the file.
        """

        self.path = path.removesuffix(".docx") + ".docx"
        self.doc = docx.Document()

        # set the font to Calibri
        self.doc.styles["Normal"].font.name = "Calibri"

    def __enter__(self):
        return self

    def __exit__(self, *args, **kwargs):
        self.doc.save(self.path)

    def add_xyz(self, obj: str or dict, title: str):
        """
        Add the coordinates and information about a calculation to the SI.
        It will add the electronic bond energy, Gibb's free energy, enthalpy and imaginary mode, as well as the coordinates of the molecule.

        Args:
                obj: a string specifying a calculation directory or a `TCutility.results.Result` object from a calculation.
                title: title to be written before the coordinates and information.
        """
        if isinstance(obj, str):
            obj = results.read(obj)

        # title is always bold
        s = f"<b>{title}</b><br>"

        parser = HtmlToDocx()

        # add electronic energy. E should be bold and italics. Unit will be kcal mol^-1
        E = str(round(obj.properties.energy.bond, 1)).replace("-", "—")
        s += f"<b><i>E</i></b> = {E} kcal mol<sup>—1</sup><br>"

        # add Gibbs and enthalpy if we have them
        if obj.properties.energy.gibbs:
            G = str(round(obj.properties.energy.gibbs, 1)).replace("-", "—")
            s += f"<b><i>G</i></b> = {G} kcal mol<sup>—1</sup><br>"
        if obj.properties.energy.enthalpy:
            H = str(round(obj.properties.energy.enthalpy, 1)).replace("-", "—")
            s += f"<b><i>H</i></b> = {H} kcal mol<sup>—1</sup><br>"

        # add imaginary frequency if we have one
        if obj.properties.vibrations:
            if obj.properties.vibrations.number_of_imag_modes == 1:
                freq = abs(round(obj.properties.vibrations.frequencies[0]))
                s += f"<b><i>ν<sub>imag</sub></i></b> = {freq}<i>i</i> cm<sup>—1</sup>"

        # remove trailing line breaks
        s = s.removesuffix("<br>")

        # coords should be written in mono-type font with 8 decimals and 4 spaces between each coordinate
        s += "<pre>"
        for atom in obj.molecule.output:
            s += f"{atom.symbol:2}    {atom.coords[0]: .8f}    {atom.coords[1]: .8f}    {atom.coords[2]: .8f}<br>"
        s += "</pre>"
        parser.add_html_to_document(s, self.doc)

    def add_heading(self, text: str, level: int = 1):
        """
        Add a heading to the file. This method has the same arguments and functionality as docx.Document.add_heading.
        """
        self.doc.add_heading(text, level)



"""Module containing functions for quickly submitting geometry optimization jobs via the command line"""

import os

import click
import tcutility
from tcutility import job as tcjob


@click.command()
@click.option(
    "-l",
    "--level",
    type=str,
    help='Set the level of theory for the optimization. For example, "GFN1-xTB" or "BLYP-D3(BJ)/TZ2P" Can be set in the xyz-file with the \'level_of_theory\' flag.',
    default=None,
)
@click.option("-c", "--charge", type=int, help="The charge of the system. Can be set in the xyz-file with the 'charge' flag.", default=None)
@click.option("-s", "--spinpol", type=int, help="The spin-polarization of the system. Can be set in the xyz-file with the 'spinpol' flag.", default=None)
@click.option("-o", "--output", type=str, help=r"The file to write the optimized result to. By default will be written to '{xyzfile}_optimized.xyz'.", default=None)
@click.option("-k", "--keep", is_flag=True, help="Whether to keep the calculation directory after finishing the calculation.")
@click.argument("xyzfile", type=str)
def optimize_geometry(level: str, charge: int, spinpol: int, output: str, keep: bool, xyzfile: str) -> None:
    """Set up and run a geometry optimization on a given structure."""
    # set a default output
    if output is None:
        output = xyzfile.removesuffix(".xyz") + "_optimized.xyz"

    mol = tcutility.molecule.load(xyzfile)

    # get the correct job class object and set the level-of-theory
    level = level or mol.flags.level_of_theory or "GFN1-xTB"  # type: ignore # mol flags can be any
    if level.lower() == "gfn1-xtb":
        job = tcjob.DFTBJob(delete_on_finish=not keep)  # by default DFTBJob uses GFN1-xTB
    else:
        # if we are not using GFN1-xTB we will use ADF
        parts = level.split("/")
        job = tcjob.ADFJob(delete_on_finish=not keep)
        job.functional(parts[0])
        if len(parts) > 1:
            job.basis_set(parts[1])

    job.molecule(mol)
    job.optimization()
    job.charge(charge or mol.flags.charge or 0)  # type: ignore # mol flags can be any
    if isinstance(job, tcjob.ADFJob):
        job.spin_polarization(spinpol or mol.flags.spinpol or 0)  # type: ignore # mol flags can be any

    job.rundir = os.path.split(xyzfile)[0]
    job.name = f".tmp.{os.getpid()}"

    # copy the optimized mol to the output file
    job.add_postamble(f"cp {job.output_mol_path} {output}")

    job.run()



available_solvents = [
	'Vacuum',
	'AceticAcid',
	'Acetone',
	'Acetonitrile',
	'Ammonia',
	'Aniline',
	'Benzene',
	'BenzylAlcohol',
	'Bromoform',
	'Butanol',
	'isoButanol',
	'tertButanol',
	'CarbonDisulfide',
	'CarbonTetrachloride',
	'Chloroform',
	'Cyclohexane',
	'Cyclohexanone',
	'Dichlorobenzene',
	'DiethylEther',
	'Dioxane',
	'DMFA',
	'DMSO',
	'Ethanol',
	'EthylAcetate',
	'Dichloroethane',
	'EthyleneGlycol',
	'Formamide',
	'FormicAcid',
	'Glycerol',
	'HexamethylPhosphoramide',
	'Hexane',
	'Hydrazine',
	'Methanol',
	'MethylEthylKetone',
	'Dichloromethane',
	'Methylformamide',
	'Methypyrrolidinone',
	'Nitrobenzene',
	'Nitrogen',
	'Nitromethane',
	'PhosphorylChloride',
	'IsoPropanol',
	'Pyridine',
	'Sulfolane',
	'Tetrahydrofuran',
	'Toluene',
	'Triethylamine',
	'TrifluoroaceticAcid',
	'Water',
	'CH3COOH',
	'CH3COCH3',
	'CH3CN',
	'NH3',
	'C6H5NH2',
	'C6H6',
	'C6H5CH2OH',
	'CHBr3',
	'C4H9OH',
	'(CH3)2CHCH2OH',
	'(CH3)3COH',
	'CS2',
	'CCl4',
	'CHCl3',
	'C6H12',
	'C6H10O',
	'C6H4Cl2',
	'(CH3CH2)2O',
	'C4H8O2',
	'(CH3)2NCHO',
	'(CH3)2SO',
	'CH3CH2OH',
	'CH3COOCH2CH3',
	'ClCH2CH2Cl',
	'HOCH2CH2OH',
	'HCONH2',
	'HCOOH',
	'C3H8O3',
	'C6H18N3OP',
	'C6H14',
	'N2H4',
	'CH3OH',
	'CH3CH2COCH3',
	'CH2Cl2',
	'HCONHCH3',
	'C5H9NO',
	'C6H5NO2',
	'N2',
	'CH3NO2',
	'POCl3',
	'(CH3)2CHOH',
	'C5H5N',
	'C4H8SO2',
	'C4H8O',
	'C6H5CH3',
	'(CH3CH2)3N',
	'CF3COOH',
	'H2O'
]
